import errno
import json
import hashlib
import logging
from socket import error as sockerror
from time import strftime

import paramiko

from host import ssh_key_digest

jsonrpc_methods = dict()
import frontends.jsonrpc_ubnt

class QVizJSONRPC:
	"""JSONRPC frontend to QViz.
	"""

	def listensock(self, gaituple):
		from os import chmod
		from socket import socket
		from stat import S_IREAD, S_IWRITE, S_IRGRP, S_IWGRP, S_IROTH, S_IWOTH
		try:	from socket import AF_UNIX
		except ImportError:
			AF_UNIX = -1

		mode = S_IREAD|S_IWRITE|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH

		(af, st, p, cn, sa) = gaituple
		s = socket(af, st)
		s.setblocking(False)
		try:
			s.bind(sa)
		except sockerror as se:
			if af == AF_UNIX and se.errno == errno.EADDRINUSE:
				from os import unlink
				unlink(sa)
				s.bind(sa)
			else:
				raise
		finally:
			if af == AF_UNIX:
				chmod(sa, mode)
		s.listen(3)
		return s

	def __init__(self, qviz):
		from sys import argv, exit

		listeners = []
		try:
			if '-u' in argv:
				from socket import AF_UNIX, SOCK_STREAM
				i = argv.index("-u")
				listeners += [(AF_UNIX, SOCK_STREAM, 0, '', argv[i+1])]
			if '-i' in argv:
				from socket import getaddrinfo, SOCK_STREAM, SOL_TCP
				i = argv.index("-i")
				gai = getaddrinfo(argv[i+1], argv[i+2], 0, SOCK_STREAM, SOL_TCP)
				listeners += gai
		except Exception as e:
			print("Usage: qviz.py (-u socket|-i localhost 9090)")
			print(e)
			exit(1)

		try:
			if len(listeners) == 0:
				raise Exception("no arguments; UNIX socket or TCP/IP serv/name")
			self.listeners = map(self.listensock, listeners)
		except Exception as e:
			print("Usage: qviz.py (-u socket|-i localhost 9090)")
			print(e)
			exit(1)

		self.nconns = 0
		self.connections = dict()

		self.qviz = qviz
		self.qviz.frontend_key_policy = MissingKeyPolicyJSONRPC()
		self.qviz.frontend_key_policy.frontend = self
		self.wanna_run = True

		# Rebind all external methods to ourself.  We are a singleton.
		for key in jsonrpc_methods:
			setattr(QVizJSONRPC, key, jsonrpc_methods[key])
			jsonrpc_methods[key] = getattr(self, key)

		jsonrpc_methods['host_add'] = self.host_add
		jsonrpc_methods['host_delete'] = self.host_delete
		jsonrpc_methods['host_list'] = self.hosts
		jsonrpc_methods['plan_set'] = self.plan_set
		jsonrpc_methods['plan_unset'] = self.plan_unset
		jsonrpc_methods['plan_list'] = self.plans

		jsonrpc_methods['quarantine_accept'] = self.quarantine_accept
		jsonrpc_methods['quarantine_reject'] = self.quarantine_reject
		jsonrpc_methods['quarantine_list'] = self.quarantine_list

		self.ltv = dict()
		jsonrpc_methods['ltv_list'] = self.ltv_list
		jsonrpc_methods['ltv_fetch'] = self.ltv_fetch
		jsonrpc_methods['dict_get'] = self.dict_get

		logfile = '-{0}.log'.format(strftime("%Y-%m-%d"))
		self.logger = logging.getLogger('qviz')
		self.logger.setLevel(logging.INFO)

		self.logfile = open('qviz'+logfile, 'a+')
		lform = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
		lh = logging.StreamHandler(self.logfile)
		lh.setLevel(logging.INFO)
		lh.setFormatter(lform)
		self.logger.addHandler(lh)

		paramiko.util.log_to_file('paramiko'+logfile)

	def daemonize(self):
		from os import dup2, fork, setsid, umask
		pid = fork()
		if pid > 0:
			from sys import exit
			exit(0)
		setsid()
		umask(0)
		from sys import stdin, stdout, stderr
		for f in [stdout, stderr]:
			f.flush()
			dup2(self.logfile.fileno(), f.fileno())
		stdin.close()

	def log_host_status(self, h):
		self.log("{0}: {1}".format(h.name, h.status))

	def log(self, m, setbusy=False, setready=False):
		self.logger.info(m)

	def run(self):
		listen = [(x, self.accept_connection) for x in self.listeners]
		while self.wanna_run:
			l = [(x, self.handle_client) for x in self.connections]
			self.qviz.process_events(r=listen+l)

	def quit(self, dumpthis=None):
		self.wanna_run = False

	def accept_connection(self, nuttin, sock):
		self.nconns += 1
		(conn, addr) = sock.accept()
		if addr == '':
			addr = str(self.nconns)
		self.log("new connection from {0}".format(addr))
		self.connections[addr] = conn

	def handle_client(self, addr, sock=None):
		sock = self.connections[addr]
		try:
			request = sock.recv(4096)
		except sockerror as se:
			self.logger.error(se)
			request = None
		# Pipe closed.
		if not request:
			sock.close()
			self.connections[addr] = None
			del self.connections[addr]
			return
		for l in request.split('\n'):
			if not l:
				break
			try:
				self.jsonrpc(addr, l)
			except sockerror as se:
				break

	def error(self, code, message, data=None):
		e = dict()
		e['code'] = code
		e['message'] = message
		if data:
			e['data'] = data
		return json.dumps(e)

	def respond_err(self, addr, code, msg, id, data=None):
		s = '{"error": ' + self.error(code, msg, data)
		s += ', "id": ' +id+ '}'
		self.respond_to_client(addr, s)

	def respond_to_client(self, addr, msg):
		sock = self.connections[addr]
		try:
			sock.setblocking(True)
			sock.sendall(msg)
			sock.setblocking(False)
		except sockerror as se:
			self.logger.error(se)
			sock.close()
			self.connections[addr] = None
			del self.connections[addr]
			raise

	def jsonrpc(self, addr, req):
		try:
			j = json.loads(req)
		except:
			return self.respond_err(addr, -32700, "Parse error.", 'null', req)

		if 'id' not in j or 'method' not in j:
			return self.respond_err(addr, -32600, "Invalid request.", 'null')

		id = str(j['id'])
		m = str(j['method'])
		p = j['params'] if 'params' in j else None

		if not id:
			return self.respond_err(addr, -32600, "Invalid id.", 'null')

		if m not in jsonrpc_methods:
			return self.respond_err(addr, -32601, "Method not found.", id, m)
		return jsonrpc_methods[m](addr, id, p)

	# --------------------------------------------- 8>< =========
	def host_add(self, addr, id, p=None):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'hostname' not in p or 'user' not in p or 'method' not in p or not p['hostname']:
			return self.respond_err(addr, -32602, "Invalid params.", id)
		name = p['name'] if 'name' in p else p['hostname']
		if name in self.qviz.hosts:
			return self.respond_err(addr, 1, "Host already present.", id)

		nparams = dict()
		nparams['name'] = name
		nparams['hostname'] = p['hostname']
		nparams['user'] = p['user']
		if p['method'] == 'password' and 'password' in p:
			nparams['password'] = p['password']
		elif p['method'] == 'pubkey' and 'pubkey' in p:
			nparams['pubkey'] = p['pubkey']
		else:
			return self.respond_err(addr, -32602, "Invalid method.", id)
		if 'port' in p:
			nparams['port'] = int(p['port'])
		self.qviz.add_host(nparams)

		s = '{"result": "added ' + name + '", "id": '+id+' }'
		self.respond_to_client(addr, s)

	def host_delete(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'name' not in p:
			return self.respond_err(addr, -32602, "Invalid params.", id)
		h = p['name']
		if not h in self.qviz.hosts:
			return self.respond_err(addr, 2, "Host not present.", id)

		self.qviz.disconnect_host(h)

		s = '{"result": "deleted ' + h + '", "id": '+id+' }'
		self.respond_to_client(addr, s)

	def hosts(self, addr, id, p=None):
		"""RPC method."""
		todump = dict()
		for host in sorted(self.qviz.hosts):
			h = self.qviz.hosts[host]
			export = dict()
			for a in ['started', 'connected', 'went_bad']:
				export[a] = getattr(h, a)

			export['plans'] = dict()
			for p in h.plans:
				json_plan = json.loads(h.plans[p].to_JSON())
				export['plans'][p] = json_plan
			usw = [x for x in h.software if x.name == 'UBNT']
			if len(usw):
				export['status'] = usw[0].status
				export['stations'] = usw[0].stations
			todump[host] = export

		s = '{"result": ' + json.dumps(todump) + ', "id": '+id+' }'
		self.respond_to_client(addr, s)

	def plan_set(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)

		if 'host' not in p or 'subtree' not in p or 'secs' not in p:
			return self.respond_err(addr, -32602, "Invalid params.")

		(host, subtree, secs) = (p['host'], p['subtree'], p['secs'])
		if not host in self.qviz.hosts:
			return self.respond_err(addr, 2, "Host not present.", id)

		h = self.qviz.hosts[host]
		try:	h.set_plan(subtree, int(secs))
		except ValueError:
			return self.respond_err(addr, -32602, "Invalid 'seconds'.")

		s = '{"result": "plan for '+host+"'s "+subtree
		s += ' has been set", "id": '+id+' }'
		self.respond_to_client(addr, s)

	def plan_unset(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)

		if 'host' not in p or 'subtree' not in p:
			return self.respond_err(addr, -32602, "Invalid params.")

		(host, subtree) = (p['host'], p['subtree'])
		if not host in self.qviz.hosts:
			return self.respond_err(addr, 2, "Host not present.", id)

		h = self.qviz.hosts[host]
		try:	unset = h.unset_plan(subtree)
		except:
			return self.respond_err(addr, -32602, "Couldn't unset plan.", id)
		if not unset:
			return self.respond_err(addr, -32602, "Plan wasn't set.", id)
		s = '{"result": "plan for '+host+"'s "+subtree
		s += ' has been deleted", "id": '+id+' }'
		self.respond_to_client(addr, s)

	def plans(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if not p in self.qviz.hosts:
			return self.respond_err(addr, 2, "Host not present.", id)

		jsonstr = '{'
		h = self.qviz.hosts[p]
		for plan in h.plans:
			jsonstr += "'{0}': {1}".format(plan, h.plans[plan].to_JSON())
		jsonstr += '}'

		s = '{"result": ' + jsonstr + ', "id": '+id+' }'
		self.respond_to_client(addr, s)

	def quarantine_accept(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'host' not in p or 'keytype' not in p or 'keyfp' not in p:
			return self.respond_err(addr, -32602, "Invalid params.")

		(host, keytype, yourfp) = (p['host'], p['keytype'], p['keyfp'])
		if not host in self.qviz.quarantine:
			return self.respond_err(addr, 2, "Host not quarantined.", id)

		if keytype not in ['ssh-rsa', 'ecdsa-sha2-nistp521']:
			return self.respond_err(addr, -32602, "Invalid keytype.")

		(key, config) = self.qviz.quarantine[host]
		myfp = ssh_key_digest(key)
		if myfp != yourfp:
			d = dict()
			d['yours'] = yourfp
			d['mine'] = myfp
			m = "{0} fingerprints don't match.".format(keytype)
			return self.respond_err(addr, 4444, m, id, d)

		allkeys = paramiko.HostKeys(self.qviz.known_hosts)
		allkeys.add(host, keytype, key)
		allkeys.save(self.qviz.known_hosts)
		del self.qviz.quarantine[host]
		self.qviz.add_host(config)

		s = '{"result": "reconnecting '+host+'", "id": '+id+' }'
		self.respond_to_client(addr, s)

	def quarantine_reject(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'host' not in p or 'keytype' not in p or 'keyfp' not in p:
			return self.respond_err(addr, -32602, "Invalid params.")

		(host, keyfp) = (p['host'], p['keyfp'])
		if not host in self.qviz.quarantine:
			return self.respond_err(addr, 2, "Host not quarantined.", id)

		(key, config) = self.qviz.quarantine[host]
		if ssh_key_digest(key) != keyfp:
			d = dict()
			d['yours'] = yourfp
			d['mine'] = myfp
			m = "{0} fingerprints don't match.".format(keytype)
			return self.respond_err(addr, 4444, m, id, d)

		del self.qviz.quarantine[host]
		s = '{"result": "'+host+' deleted from quarantine", "id": '+id+' }'
		self.respond_to_client(addr, s)


	def quarantine_list(self, addr, id, p=None):
		"""RPC method."""
		dump = dict()
		for x in self.qviz.quarantine:
			k = self.qviz.quarantine[x][0]
			dump[x] = (k.get_name(), ssh_key_digest(k))

		s = '{"result": ' + json.dumps(dump) + ', "id": '+id+' }'
		self.respond_to_client(addr, s)

	# --------------------------------------------- 8>< =========
	def ask_for_password(self, message=None):
		"""The daemon doesn't have any means of asking for passwords."""
		return None

	def add_host(self, host):
		"""Called by qviz main class to give it a fake widget."""
		key = str(host)
		self.ltv[key] = ("host", host, dict())
		host.widget = [key]
		return key

	def ltv_add_node(self, parent, desc, t, id):
		"""Klicem je @id."""
		assert type(parent) == list
		cur = self.ltv
		for level in parent:
			if level not in cur:
				cur[level] = (None, None, dict())
			cur = cur[level][2]
		cur[id] = (t, id, dict())
		return parent + [id]

	def ltv_delete_node(self, path):
		assert type(path) == list
		cur = self.ltv
		for level in path:
			if level == path[-1]:
				cur.pop(level)
			if level not in cur:
				return
			cur = cur[level][2]

	def ltv_update_node(self, w, d, t, id):
		"""Klicem je @id."""
		assert type(w) == list
		cur = self.ltv
		for level in w:
			if level == w[-1]:
				if level in cur:
					children = cur[level][2]
					cur.pop(level)
					cur[id] = (t, id, children)
				else:
					cur[id] = (t, id, dict())
				return
			assert level in cur
			cur = cur[level][2]

	def ltv_fetch(self, addr, id, p=None):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'path' not in p:
			return self.respond_err(addr, -32602, "Invalid params.", id)
		cur = self.ltv
		for level in p['path']:
			if level not in cur:
				return self.respond_err(addr, 2, "Path not present.", id)
			cur = cur[level][2]
		s = '{"result": ' + json.dumps(cur) + ', "id": '+id+' }'
		self.respond_to_client(addr, s)

	def ltv_list(self, addr, id, p):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'path' not in p:
			return self.respond_err(addr, -32602, "Invalid params.")
		cur = self.ltv
		for level in p['path']:
			if level not in cur:
				return self.respond_err(addr, 2, "Path not present.", id)
			cur = cur[level][2]
		s = '{"result": ' + json.dumps(cur.keys()) + ', "id": '+id+' }'
		self.respond_to_client(addr, s)

	def dict_get(self, addr, id, p=None):
		"""RPC method."""
		if not p:
			return self.respond_err(addr, -32602, "No parameters.", id)
		if 'path' not in p:
			return self.respond_err(addr, -32602, "Invalid params.")
		if p['path'][0] not in self.qviz.hosts:
			return self.respond_err(addr, -32602, "No such host.")
		host = self.qviz.hosts[p['path'][0]]
		cur = self.ltv
		npath = []
		for level in p['path']:
			if level not in cur:
				return self.respond_err(addr, 2, "Path not present.", id)
			npath += [(cur[level][0], cur[level][1])]
			cur = cur[level][2]

		(t, h) = host.whatis(npath)
		s = '{"result": ' + json.dumps(h) + ', "id": '+id+' }'
		self.respond_to_client(addr, s)

	def ltv_refresh(self, widget, force=False):
		"""Any idea what this can do?  Logging?"""
		pass

class MissingKeyPolicyJSONRPC(paramiko.WarningPolicy):
	"""SSH key policy which needs confirmation by a special RPC."""
	def missing_host_key(self, client, host, key):
		keys = client.get_host_keys()
		expected = keys[host] if host in keys else None

		# replicate the adding process with the new key in place
		if host in self.frontend.qviz.hosts:
			h = self.frontend.qviz.hosts[host]
		else:
			all = self.frontend.qviz.hosts
			h = [x for x in all if all[x].hostname == host]
			h = all[h[0]]
		config = dict()
		config['hostname'] = host
		config['name'] = h.name
		config['port'] = h.port
		config['user'] = h.username
		config['password'] = h.password
		self.frontend.qviz.quarantine[host] = (key, config)

		raise paramiko.BadHostKeyException(h, key, expected)

class current_frontend(QVizJSONRPC):
	pass
