# -*- coding:utf-8 -*-

#
# Functions to return Ubiquiti stations and status information over JSON RPC.
# See bottom for type mapping.  These mappings are rebound for QVizJSONRPC's
# singleton instance in its ctor, in order for 'self' to work :-)
#
import json

def ubnt_find_stations(self, addr, id, p=[]):
	(done, max) = (0, len(p))
	todump = dict()
	for hostname in self.qviz.hosts:
		h = self.qviz.hosts[hostname]
		usw = [x for x in h.software if x.name == 'UBNT']
		if len(usw) == 0:
			continue
		sw = usw[0]
		for mac in sw.stations:
			if mac not in p:
				continue
			todump[mac] = sw.stations[mac]
			done += 1
		if done >= max:
			break
	if done == 0:
		self.respond_err(addr, 42, "No such stations.", id)
	else:
		s = '{"result": ' +json.dumps(todump)+ ', "id": '+id+' }'
		self.respond_to_client(addr, s)

def ubnt_list_stations(self, addr, id, p):
	if not p:
		return self.respond_err(addr, -32602, "No parameters.", id)
	if 'session' not in p:
		return self.respond_err(addr, -32602, "Invalid parameters.", id)
	hostname = p['session']
	if not hostname in self.qviz.hosts:
		return self.respond_err(addr, 2, "Host not present.", id)
	h = self.qviz.hosts[hostname]
	usw = [x for x in h.software if x.name == 'UBNT']
	if len(usw) == 0:
		return self.respond_err(addr, 2, "Host not UBNT.", id)

	sw = usw[0]
	jsonstr = json.dumps(sw.stations.keys())
	s = '{"result": ' + jsonstr + ', "id": '+id+' }'
	self.respond_to_client(addr, s)

def ubnt_status(self, addr, id, p):
	if not p:
		return self.respond_err(addr, -32602, "No parameters.", id)
	if 'session' not in p:
		return self.respond_err(addr, -32602, "Invalid parameters.", id)
	session = p['session']
	if not session in self.qviz.hosts:
		return self.respond_err(addr, 2, "Host not present.", id)

	h = self.qviz.hosts[session]
	usw = [x for x in h.software if x.name == 'UBNT']
	if len(usw) == 0:
		return self.respond_err(addr, 2, "Host not UBNT.", id)

	s = '{"result": ' + json.dumps(usw[0].status) + ', "id": '+id+' }'
	self.respond_to_client(addr, s)


def ubnt_firmware_list(self, addr, id, p=None):
	from os import listdir
	from hashlib import md5

	fw = listdir("firmware")
	out = dict()
	for f in fw:
		m = md5(open("firmware/"+f).read())
		out[f] = m.hexdigest()

	s = '{"result": ' + json.dumps(out) + ', "id": '+id+' }'
	self.respond_to_client(addr, s)

def ubnt_firmware_upgrade(self, addr, id, p):
	if not p:
		return self.respond_err(addr, -32602, "No parameters.", id)
	if 'session' not in p or 'file' not in p:
		return self.respond_err(addr, -32602, "Invalid params.")

	(session, file) = (p['session'], p['file'])
	if not session in self.qviz.hosts:
		return self.respond_err(addr, 2, "Host not present.", id)

	h = self.qviz.hosts[session]
	usw = [x for x in h.software if x.name == 'UBNT']
	if len(usw) == 0:
		return self.respond_err(addr, 2, "Host not UBNT.", id)
	usw[0].async_upgrade_firmware(file)

	s = '{"result": "firmware upgrade started", "id": '+id+' }'
	self.respond_to_client(addr, s)

# ---------------------------------------- 8>< ===================
#  attach the functions, so the frontend can see them

from frontends.jsonrpc import jsonrpc_methods
jsonrpc_methods['ubnt_status'] = ubnt_status
jsonrpc_methods['ubnt_find_stations'] = ubnt_find_stations
jsonrpc_methods['ubnt_list_stations'] = ubnt_list_stations
jsonrpc_methods['ubnt_firmware_list'] = ubnt_firmware_list
jsonrpc_methods['ubnt_firmware_upgrade'] = ubnt_firmware_upgrade

