# -*- coding: utf-8 -*-

from time import strftime
from getpass import getuser

# Python 2 vs 3
try:
	from Tkinter import N, S, W, E
	import Tkinter
	import tkFont
	import tkSimpleDialog
	import tkMessageBox
	import ttk
except ImportError:
	from tkinter import N, S, W, E, ttk
	import tkinter as Tkinter
	from tkinter import font as tkFont
	from tkinter import simpledialog as tkSimpleDialog
	from tkinter import messagebox as tkMessageBox

from paramiko import SSHConfig, WarningPolicy, HostKeys
from parsers.tc import TcFilter

tkinter_type2fn = dict()
import frontends.tkinter_pf
import frontends.tkinter_tc

class QVizTkinter:
	"""Tkinter frontend to QViz.
	"""

	def init_main_frame(self):
		pw = self.panedwin_lr = Tkinter.PanedWindow(self.win)

		# left-right PanedWindow -- left part, frame with treeview
		f = Tkinter.Frame(pw)
		s = Tkinter.Scrollbar(f, orient=Tkinter.VERTICAL)
		self.ltv = ttk.Treeview(f, yscrollcommand=s.set)
		self.ltv.bind('<<TreeviewSelect>>', self.__cmd_ltv_select)
		s.config(command=self.ltv.yview)

		self.ltv.heading("#0#", text="connected hosts")
		self.ltv.pack(fill=Tkinter.BOTH, expand=True, side=Tkinter.LEFT)
		s.pack(fill=Tkinter.Y, side=Tkinter.RIGHT)
		f.pack(fill=Tkinter.BOTH, expand=True)
		pw.add(f, width=self.totw*.25)

		# left-right PanedWindow -- right part, up-down PanedWindow
		pwub = self.panedwin_ub = Tkinter.PanedWindow(pw, \
		    orient=Tkinter.VERTICAL)

		# up-down PanedWindow -- up part, Frame with status+stuff
		f = Tkinter.Frame(pwub)
		self.status = Tkinter.Label(f, text='READY', foreground='yellow',
		    background='darkgreen')
		self.status.config(font=tkFont.Font(weight='bold'))
		self.status.pack(fill=Tkinter.X, expand=False, anchor=N)

		self._stuff = Tkinter.Frame(f)
		self._stuff.pack(fill=Tkinter.BOTH, expand=True)

		# this widget will be replaced a lot, as it contains user shit
		self.stuff = Tkinter.Frame(self._stuff)
		self.stuff.pack(fill=Tkinter.BOTH, expand=True)
		self.stuff.config(height=300)
		f.pack(fill=Tkinter.BOTH, expand=True)
		pwub.add(f)

		# up-down PanedWindow -- down part, Listbox with events
		self.events = Tkinter.Listbox(self.panedwin_ub)
		self.events.pack(fill=Tkinter.BOTH, expand=True, anchor=S)
		pwub.add(self.events)

		pwub.pack(fill=Tkinter.BOTH, expand=True)

		# pack the outer, left-right PanedWindow
		pw.add(pwub)
		pw.pack(fill=Tkinter.BOTH, expand=True)
		
	def init_main_menu(self):
		self.main_menu.add_command(label='quit', command=self.quit, \
		    accelerator='Ctrl+Q')
		self.win.bind_all('<Control-q>', self.quit)
		self.main_menu.add_command(label='add new host', 	\
		    command=self.__cmd_add_host, accelerator='Ctrl+A')
		self.win.bind_all('<Control-a>', self.__cmd_add_host)
		self.main_menu.add_command(label='refresh', 	\
		    command=self.__cmd_refresh_host, accelerator='Ctrl+R')
		self.win.bind_all('<Control-r>', self.__cmd_refresh_host)

		tkinter_type2fn["host"] = self.display_host

	def __init__(self, qviz):
		self.qviz = qviz
		self.qviz.frontend_key_policy = MissingKeyPolicyTkinter()
		self.qviz.frontend_key_policy.frontend = self

		self.win = Tkinter.Tk()
		self.win.rowconfigure(0, weight=1)
		self.win.columnconfigure(0, weight=1)

		# This is for windowed managers only.
		self.totw = self.win.winfo_screenwidth()
		self.toth = self.win.winfo_screenheight()
		w = int(self.totw * .8)
		h = int(self.toth * .7)
		x = int(self.totw * .1)
		y = int(self.toth * .1)
		self.win.geometry("{0}x{1}+{2}+{3}".format(w, h, x, y))

		self.main_menu = Tkinter.Menu(self.win)
		self.init_main_menu()
		self.win.config(menu=self.main_menu)

		# Whether the user wishes (or needs) to use sudo(8) all the time.
		# Controlled by the checkbox in display_host()
		self.use_sudo = Tkinter.IntVar()

		self.wanna_run = True
		self.x = 0
		self.init_main_frame()


	def ready(self):
		self.status.config(text="READY", background='darkgreen', \
		    foreground='yellow')
		self.status.update()

	def busy(self, text=None):
		text = "BUSY" + ((" with " + text) if text else "")
		self.status.config(text=text, background='red', 	\
		    foreground='white')
		self.status.update()

	def log_host_status(self, h):
		self.log("{0}: {1}".format(h.name, h.status))

	def log(self, m, setbusy=False, setready=False):
		if setbusy and not setready:
			self.busy()
		if setready and not setbusy:
			self.ready()
		if m:
			t = strftime("%Y-%m-%d %H:%M:%S")
			self.events.insert(0, "{0}: {1}".format(t, m))
			self.events.update()

	def show_error(self, title, m):
		tkMessageBox.showerror(title, m)

	def add_host(self, host):
		"""Needs to be in frontend, we don't have a unified root."""
		self.ready()
		self.log_host_status(host)

		tags = ["host", str(host)]
		host.widget = self.ltv.insert('', Tkinter.END, text=host, tags=tags)
		self.ltv_refresh(host.widget, force=True)
		self.ltv.item(host.widget, open=True)
		return host.widget

	def ltv_add_node(self, parent, text, t, id):
		tags = [t, id]
		return self.ltv.insert(parent, Tkinter.END, text=text, tags=tags)

	def ltv_delete_node(self, node):
		for x in self.ltv.get_children(node):
			self.ltv.delete(x)
		self.ltv.delete(node)

	def ltv_update_node(self, node, text, t, id):
		tags = [t, id]
		self.ltv.item(node)['text'] = text
		self.ltv.item(node)['tags'] = tags

	def ltv_refresh(self, widget, force=False):
		if force:
			self.ltv.focus(widget)
			self.ltv.see(widget)
			self.ltv.selection_set(widget)
		s = self.ltv.selection()
		if s and widget in s:
			self.ltv.event_generate("<<TreeviewSelect>>")

	def ask_for_password(self, prompt):
		d = PasswordDialog(self.win, prompt)
		return d.password

	def __cmd_add_host(self, dumpthis=None):
		d = NewConnectionDialog(self.win)
		if d.result and 'hostname' in d.result and d.result['hostname']:
			self.busy()
			self.qviz.add_host(d.result)
			self.ltv.focus_set()

	def __cmd_refresh_host(self, dumpthis=None):
		n = self.ltv.focus()
		# find the root
		while self.ltv.parent(n):
			n = self.ltv.parent(n)
		i = self.ltv.item(n)
		[_, hostname] = i['tags']
		self.refresh_host(hostname, use_sudo=self.use_sudo.get())
		self.ltv.focus_set()

	def __cmd_ltv_select(self, ev):
		tree = ev.widget
		node = tree.focus()

		if 'tags' not in tree.item(node):
			print("well thats bad")
			return
		o = tree.item(node)['tags']
		# save the tree path and find host
		n = node
		path = [tree.item(n)['tags']]
		while tree.parent(n):
			n = tree.parent(n)
			i = tree.item(n)
			path = path + [i['tags']]
		path.reverse()
		if path[0][0] != 'host':
			print("path not starting with host, bad")
			return
		if path[0][1] not in self.qviz.hosts:
			print("host disconnected, but failed to disappear")
			self.ltv_delete_node(node)
			return

		host = self.qviz.hosts[path[0][1]]
		# The host will tell us what do we need to draw,
		# for example ["filter", "basic", "meta"]
		# and each frontend uses this to find the
		# best function to draw the content.

		try:
			(t, h) = host.whatis(path)
		except:
			[t, v] = path[-1]
			self.log("{0} {1} is gone.".format(t, v))
			return
		fn = self.visualize_type(t)
		self.stuff.pack_forget()
		self.stuff.destroy()
		self.stuff = Tkinter.Frame(self._stuff)
		fn(self, h, path)
		self.stuff.pack(fill=Tkinter.BOTH, expand=True)

		#tkMessageBox.showinfo('blem', str(dir(ev.widget)))

	def run(self):
		while self.wanna_run:
			self.win.update()
			self.qviz.process_events()

	def quit(self, dumpthis=None):
		self.wanna_run = False

	def visualize_type(self, t):
		"""Based on the type of the object, find out what can I draw."""
		k = ' '.join(t)
		if k in tkinter_type2fn:
			return tkinter_type2fn[k]

		# try worse match (Array[:-0] is empty, that's why!)
		for i in range(1, len(t)):
			k = " ".join(t[:-i])
			if k in tkinter_type2fn:
				return tkinter_type2fn[k]
			print("{0} not available to draw".format(k))
		return self.display

	def display_dict_table(self, parent, table):
		forbid = ['widget', 'qdcl', 'qdiscs', 'classes', 'filters', \
		    'ingress', 'egress', 'parent']
		def yview(*args):
			apply(lbk.yview, args)
			apply(lbv.yview, args)
		f = Tkinter.Frame(parent)
		f.rowconfigure(0, weight=1)
		f.columnconfigure(0, weight=2)
		f.columnconfigure(1, weight=1)
		f.columnconfigure(2, weight=32)
		scroll = Tkinter.Scrollbar(f, orient=Tkinter.VERTICAL)
		lbk = Tkinter.Listbox(f, yscrollcommand=scroll.set)
		lbv = Tkinter.Listbox(f, yscrollcommand=scroll.set)
		scroll.config(command=yview)

		for k in sorted(table):
			if k in forbid:
				continue
			lbk.insert(Tkinter.END, str(k))
			lbv.insert(Tkinter.END, str(table[k]))

		lbk.grid(row=0, column=0, sticky=N+S+W+E)
		lbv.grid(row=0, column=2, sticky=N+S+W+E)
		scroll.grid(row=0, column=1, sticky=N+S)
		return f

	def display(self, notused, handle, path):
		[key, name] = path[-1]

		m = "No specialized widget for {0} {1}".format(key, name)
		l = Tkinter.Label(self.stuff, text=m, fg='yellow', bg='red')
		l.config(font=tkFont.Font(weight='bold'))
		l.pack(fill=Tkinter.X, expand=False)

		m = "Tree path: "
		for [t, n] in path:
			m += "({0}: {1}), ".format(t, n)
		l = Tkinter.Label(self.stuff, text=m[:-2])
		l.pack(fill=Tkinter.X, expand=False)

		if isinstance(handle, dict):
			f = self.display_dict_table(self.stuff, handle)
			f.pack(fill=Tkinter.BOTH, expand=True)

	def display_host(self, notused, handle, path):
		[key, name] = path[-1]

		c = Tkinter.Checkbutton(self.stuff, text="Use sudo", \
		    variable=self.use_sudo)
		c.pack(fill=Tkinter.X, expand=False)

		m = "Refresh host {0}".format(handle.name)
		b = Tkinter.Button(self.stuff, text=m)
		b.config(font=tkFont.Font(weight='bold'))
		b.config(command=lambda: self.refresh_host(handle.name, \
		    use_sudo=self.use_sudo.get()))
		b.pack(fill=Tkinter.X, expand=False)

		m = "Disconnect from {0}".format(handle.name)
		b = Tkinter.Button(self.stuff, text=m)
		b.config(font=tkFont.Font(weight='bold'))
		b.config(command=lambda: self.disconnect_host(handle.name))
		b.pack(fill=Tkinter.X, expand=False)

		m = "Tree path: "
		for [t, n] in path:
			m += "({0}: {1}), ".format(t, n)
		l = Tkinter.Label(self.stuff, text=m[:-2])
		l.pack(fill=Tkinter.BOTH, expand=True)

		# separate frame for plans
		plans = Tkinter.Frame(self.stuff)

		p_top = Tkinter.Label(plans, text='Periodic actions')
		p_top.config(font=tkFont.Font(weight='bold'))
		p_top.grid(column=0, row=0, columnspan=3)

		l = Tkinter.Listbox(plans)
		l.grid(column=2, row=1, rowspan=4, sticky=N+S+W+E)
		for p in handle.plans:
			m = "{0} ({1} sec.)".format(str(p), handle.plans[p].secs)
			l.insert(Tkinter.END, m)

		handle.fe_subtree = Tkinter.StringVar()
		np = ttk.Combobox(plans, state='readonly',	\
		    textvariable=handle.fe_subtree)
		np['values'] = list(handle.subtrees.keys())
		np.grid(column=0, row=1, columnspan=2, sticky=W+E)

		np_label = Tkinter.Label(plans, text='Seconds')
		np_label.grid(column=0, row=2)
		handle.fe_secs = Tkinter.IntVar()
		handle.fe_secs.set(60)
		np_secs = Tkinter.Entry(plans, textvariable=handle.fe_secs)
		np_secs.grid(column=1, row=2)

		def __add_plan():
			subtree = handle.fe_subtree.get()
			if not subtree:
				m = "No subtree was selected from the list."
				return tkMessageBox.showerror(handle, m)
			if subtree not in handle.subtrees:
				return tkMessageBox.showerror(handle, \
				    "Selected subtree no longer supported.")
			secs = handle.fe_secs.get()
			if not secs:
				return tkMessageBox.showerror(handle, "No secs.")
			handle.set_plan(subtree, secs)
			self.ltv_refresh(handle.widget)
			m = "{0}'s {1} planned each {2} seconds."
			m = m.format(str(handle), subtree, secs)
			return tkMessageBox.showinfo(handle, m)

		np_add = Tkinter.Button(plans, text='Add plan',	\
		    command=__add_plan)
		np_add.grid(column=0, row=3, columnspan=2)

		def __del_plan():
			sel = l.curselection()
			if not sel:
				m = "No plan was selected to be deleted."
				return tkMessageBox.showerror(handle, m)
			m = l.get(sel)
			subtree = m.split()[0]
			handle.unset_plan(subtree)
			self.ltv_refresh(handle.widget)
			return tkMessageBox.showinfo(handle, subtree+" not planned.")
		np_del = Tkinter.Button(plans, text='Delete plan',	\
		    command=__del_plan)
		np_del.grid(column=0, row=4, columnspan=2)

		plans.grid_columnconfigure(0, weight=1)
		plans.grid_columnconfigure(1, weight=1)
		plans.grid_columnconfigure(2, weight=2)
		plans.pack(fill=Tkinter.BOTH, expand=True)

	def disconnect_host(self, who, ui_origin=True):
		self.busy()
		self.stuff.pack_forget()
		self.stuff.destroy()
		self.stuff = Tkinter.Frame(self._stuff)
		self.stuff.pack(fill=Tkinter.BOTH, expand=True)
		if ui_origin:
			self.qviz.disconnect_host(who)
		self.ready()

	def refresh_host(self, who, use_sudo=False):
		self.busy()
		try:
			self.qviz.refresh_host(who, use_sudo)
		except Exception as e:
			self.log("{0}: {1}".format(who, str(e)))
		self.ready()

class PasswordDialog(tkSimpleDialog.Dialog):
	"""Pops up L{prompt} and asks for a password.  What'd'ya expect?"""
	def __init__(self, parent, prompt):
		self.prompt = prompt
		self.p = Tkinter.StringVar()
		self.password = None
		tkSimpleDialog.Dialog.__init__(self, parent, prompt)

	def body(self, master):
		Tkinter.Label(master, text=self.prompt).grid(row=0)
		self.pwd = Tkinter.Entry(master, show='*', textvariable=self.p)
		self.pwd.grid(row=1)
		return self.pwd

	def apply(self):
		self.password = self.p.get()

class NewConnectionDialog(tkSimpleDialog.Dialog):
	"""A dialog that asks for parameters for a new SSH connection."""
	def body(self, master):
		Tkinter.Label(master, text="Host Name").grid(row=0)
		Tkinter.Label(master, text="User Name").grid(row=1)
		Tkinter.Label(master, text="Password").grid(row=2)
		Tkinter.Label(master, text="Port").grid(row=3)
		self.hostname = Tkinter.Entry(master)
		self.username = Tkinter.Entry(master)
		self.password = Tkinter.Entry(master, show='*')
		self.port = Tkinter.Entry(master, text="22")
		self.hostname.grid(row=0, column=1)
		self.username.grid(row=1, column=1)
		self.password.grid(row=2, column=1)
		self.port.grid(row=3, column=1)

		from os.path import expanduser, join
		from os import mkdir, stat
		home = expanduser('~')
		path = join(home, '.ssh', 'config')
		try:
			stat(path)
		except:
			m = "Missing ~/.ssh/config.  Create?"
			if not tkMessageBox.askyesno('missing ' + path, m):
				return self.hostname
			p = join(home, '.ssh')
			try:	mkdir(p)
			except:	pass
			# any better touch(1) ?
			try:
				f = open(path, "w")
				f.close()
			except IOError as ioe:
				m = "Can't create {0}, {1}".format(path, ioe)
				tkMessageBox.showerror(path, m)
		
		self.c = SSHConfig()
		try:
			self.c.parse(open(path))
		except IOError as ioe:
			m = "Can't open {0}, {1}".format(path, ioe)
			tkMessageBox.showerror(path, m)

		if not hasattr(self.c, "_config"):
			m = "Paramiko has changed SSHConfig format, fix it."
			tkMessageBox.showerror(path, m)
			return hostname

		s = Tkinter.Scrollbar(master, orient=Tkinter.VERTICAL)
		self.lb = Tkinter.Listbox(master, yscrollcommand=s.set)
		s.config(command=self.lb.yview)
		for item in self.c._config:
			if item['host'][0] == '*':
				continue
			self.lb.insert(Tkinter.END, item['host'][0])
		self.lb.bind('<<ListboxSelect>>', self.onselect)
		self.lb.bind('<Double-Button-1>', self.ok)
		self.lb.grid(row=0, column=2, rowspan=4)
		s.grid(row=0, column=3, rowspan=4, sticky=Tkinter.N+Tkinter.S)
		return self.hostname # initial focus

	def onselect(self, event):
		"""Fills in textfields from ssh_config(5) selection."""
		sel = event.widget.curselection()
		h = event.widget.get(sel)
		c = self.c.lookup(h)
		self.hostname.delete(0, Tkinter.END)
		if 'hostname' in c:
			self.hostname.insert(0, c['hostname'])
		self.username.delete(0, Tkinter.END)
		if 'user' in c:
			self.username.insert(0, c['user'])
		else:
			self.username.insert(0, getuser())
		self.port.delete(0, Tkinter.END)
		if 'port' in c:
			self.port.insert(0, c['port'])
		else:
			self.port.insert(0, "22")

	def apply(self):
		h = self.hostname.get()
		u = self.username.get()
		p = self.password.get()
		port = self.port.get()

		# if the user selected something from ssh_config(5), use that
		sel = self.lb.curselection()
		if sel:
			hsel = self.lb.get(sel)
			self.result = self.c.lookup(hsel)
			# selected and not modified hostname
			if h == self.result['hostname']:
				if 'user' not in self.result or \
				    u != self.result['user']:
					self.result['user'] = u
				if u == '':
					self.result['user'] = getuser()
				self.result['password'] = p
				self.result['name'] = hsel
				self.result['port'] = port
				return

		# she decided to alter it, just use the fields provided
		self.result = dict()
		self.result['hostname'] = h
		if u == '':
			self.result['user'] = getuser()
		else:
			self.result['user'] = u
		self.result['password'] = p
		self.result['port'] = port

class MissingKeyPolicyTkinter(WarningPolicy):
	"""SSH key policy which shows a polite window for Tkinter frontend."""
	def missing_host_key(self, client, h, key):
		from host import ssh_key_digest
		d = ssh_key_digest(key)

		type = key.get_name()
		m = "MD5({2}({0}))\n{1}\nAccept the key?".format(h, d, type)
		if tkMessageBox.askyesno('missing host key', m):
			client._host_keys.add(h, key.get_name(), key)

			# Store the key to known_hosts as OpenSSH would.
			allkeys = HostKeys(self.frontend.qviz.known_hosts)
			allkeys.add(h, type, key)
			allkeys.save(self.frontend.qviz.known_hosts)
			return

		keys = client.get_host_keys()
		expected = keys[h] if h in keys else None
		
		from paramiko import BadHostKeyException
		raise BadHostKeyException(h, key, expected)

class current_frontend(QVizTkinter):
	pass
