# -*- coding: utf-8 -*-

#
# Functions to visualize tc(8) things in Tkinter.  See bottom for type mapping.
#

try:
	import Tkinter
	import tkFont
	import ttk
	from Tkinter import N, S, W, E
except ImportError:
	import tkinter as Tkinter
	from tkinter import N, S, W, E, ttk
	from tkinter import font as tkFont

from collections import deque

from parsers.tc import TcParseBPF, TcParseEmatch, TcPrintEmatch
from software import RootPriviledgesNeededException

def widget_actions(parent, acts):
	"""Returns a Listbox added to @parent, full of actions in @acts."""
	lba = Tkinter.Listbox(parent)
	for a in acts:
		t = " ".join(["action", a.type])
		fn = tkinter_type2fn[t] if t in tkinter_type2fn else str
		lba.insert(Tkinter.END, fn(a))
	return lba

def widget_bpf(parent, string):
	try:
		x = TcParseBPF(string)
	except Exception as e:
		return Tkinter.Label(parent, text=str(e))

	if not x:
		return Tkinter.Label(parent, text="No BPF bytecode!")
	lb = Tkinter.Listbox(parent)
	for insn in x:
		t = insn.split('\t')
		if len(t) == 3:
			lb.insert(Tkinter.END, t[0] +' '+ t[1])
		else:
			lb.insert(Tkinter.END, '     '+ t[0])
	return lb

def widget_classid(fe, handle, hostname, dev, parent=None):
	"""Handle has some class or qdisc it points at, and wants a button to it."""
	clid = None
	if not parent:
		parent = fe.stuff
	if hasattr(handle, 'classid'):
		host = fe.qviz.hosts[hostname]
		path_stub = [["host", hostname], ["interface", dev]]

		try:
			qpath = path_stub + [["qdisc", handle.classid]]
			(t, clid) = host.whatis(qpath)
		except:
			try:
				cpath = path_stub + [["class", handle.classid]]
				(t, clid) = host.whatis(cpath)
			except:
				pass # stays at None from the beginning
	else:
		m = "Doesn't direct traffic anywhere."
		return Tkinter.Label(parent, text=m)

	if clid:
		m = "Go to {0} {1}".format(clid['qdcl'], handle.classid)
		return Tkinter.Button(parent, text=m, \
		    command=lambda: fe.ltv_refresh(clid['widget']))
	else:
		m = "Neither qdisc nor class {0} exists!".format(handle.classid)
		return Tkinter.Label(parent, text=m)

def widget_ematch(parent, string):
	rpnemsg = "You need root priviledges to display {0}, use sudo!"
	try:
		x = TcParseEmatch(string)
	except RootPriviledgesNeededException as rpne:
		s = rpnemsg.format(str(rpne))
		return Tkinter.Label(parent, text=s)

	if not x:
		return Tkinter.Label(parent, text="No ematches!")
	(t, _) = x
	tv = ttk.Treeview(parent)
	q = deque()
	q.append(('', t))
	while len(q):
		(pw, n) = q.popleft()
		try:
			s = TcPrintEmatch(n)
		except RootPriviledgesNeededException as rpne:
			s = rpnemsg.format(str(rpne))
			tv.insert(pw, Tkinter.END, text=s, open=True)
			return tv

		w = tv.insert(pw, Tkinter.END, text=s, open=True)
		if n['type'] in ['and', 'or']:
			q.append((w, n['left']))
			q.append((w, n['right']))
	return tv

# XXX get rid
def display_qdisc(fe, handle, path):
	return fe.display(fe, handle, path)
	[key, name] = path[-1]
	m = "qdisc, {0} {1}".format(key, name)
	Tkinter.Label(fe.stuff, text=m).pack()
	Tkinter.Label(fe.stuff, text=str(path)).pack(expand=True)

def display_filter(fe, handle, path):
	[key, name] = path[-1]

	if len(handle.children) == 1:
		filter_single_child(display_filter, handle.type, fe, handle, path)
		return
	m = "Filter {0}, protocol {1} preference {2}, actions: {3}"
	a = ','.join(map(str, handle.actions))
	m = m.format(handle.type, handle.proto, handle.pref, a)
	top = Tkinter.Label(fe.stuff, text=m)
	top.config(font=tkFont.Font(weight='bold'))
	top.pack()
	Tkinter.Label(fe.stuff, text=str(path)).pack(expand=True)
	Tkinter.Label(fe.stuff, text=str(handle)).pack(expand=True)

def filter_single_child(function, name, fe, handle, path):
	"""If a filter has only one child, this will display it with a warning."""

	(k, v) = handle.children.popitem()
	handle.children[k] = v
	m = "Filter {0}, protocol {1} preference {2}, "
	m += "has only one child: handle 0x{3:x}."
	m = m.format(name, handle.proto, handle.pref, k)
	l = Tkinter.Label(fe.stuff, text=m)
	l.config(font=tkFont.Font(weight='bold'))
	l.pack(fill=Tkinter.X, ipady=5)
	function(fe, v, path + [["filter", k]])

def display_filter_basic(fe, handle, path):
	[key, name] = path[-1]

	if len(handle.children) == 1:
		filter_single_child(display_filter_basic, "basic", fe, handle, path)
		return
	if not handle.handle:
		m = "basic filter protocol {0} preference {1}"
		m = m.format(handle.proto, handle.pref)
		Tkinter.Label(fe.stuff, text=m).pack()
		Tkinter.Label(fe.stuff, text="Choose parts of the filter from the tree").pack(fill=Tkinter.BOTH, expand=True)
		return

	m = "basic filter {0}".format(name)
	Tkinter.Label(fe.stuff, text=m).pack(fill=Tkinter.X)

	pw = Tkinter.PanedWindow(fe.stuff)
	pw.columnconfigure(0, weight=1)
	pw.columnconfigure(1, weight=2)

	lf = Tkinter.Frame(pw)
	rf = Tkinter.Frame(pw)
	Tkinter.Label(lf, text="Extended matches:").pack(fill=Tkinter.X)
	Tkinter.Label(rf, text="Chain of actions:").pack(fill=Tkinter.X)

	lba = widget_actions(rf, handle.actions)
	lba.pack(fill=Tkinter.BOTH, expand=True)
	emt = widget_ematch(lf, b' '.join(handle.ematch))
	emt.pack(fill=Tkinter.BOTH, expand=True)

	pw.add(lf)
	pw.add(rf)
	pw.pack(fill=Tkinter.BOTH, expand=True)

def display_filter_bpf(fe, handle, path):
	[key, name] = path[-1]

	if len(handle.children) == 1:
		filter_single_child(display_filter_bpf, "bpf", fe, handle, path)
		return
	if not hasattr(handle, 'bytecode'):
		m = "bpf filter protocol {0} preference {1}"
		m = m.format(handle.proto, handle.pref)
		Tkinter.Label(fe.stuff, text=m).pack()
		Tkinter.Label(fe.stuff, text="Choose parts of the filter from the tree").pack(fill=Tkinter.BOTH, expand=True)
		return

	pw = Tkinter.PanedWindow(fe.stuff)
	pw.columnconfigure(0, weight=1)
	pw.columnconfigure(1, weight=2)

	lf = Tkinter.Frame(pw)
	rf = Tkinter.Frame(pw)
	Tkinter.Label(lf, text="BPF bytecode:").pack(fill=Tkinter.X)

	m = "Filter {0}, protocol {1} preference {2}"
	m = m.format(handle.type, handle.proto, handle.pref)
	top = Tkinter.Label(rf, text=m)
	top.config(font=tkFont.Font(weight='bold'))
	top.pack(fill=Tkinter.X)
	Tkinter.Label(rf, text=str(handle)).pack(fill=Tkinter.X)

	Tkinter.Label(rf, text="Chain of actions:").pack(fill=Tkinter.X)

	lba = widget_actions(rf, handle.actions)
	lba.pack(fill=Tkinter.BOTH, expand=True)
	emt = widget_bpf(lf, handle.bytecode)
	emt.pack(fill=Tkinter.BOTH, expand=True)
	bcode = Tkinter.Text(rf, height=1)
	bcode.insert(1.0, "BPF bytecode: " + handle.bytecode.decode('ascii'))
	bcode.pack(fill=Tkinter.X)

	pw.add(lf)
	pw.add(rf)
	pw.pack(fill=Tkinter.BOTH, expand=True)

def display_filter_flow(fe, handle, path):
	[key, name] = path[-1]

	if len(handle.children) == 1:
		filter_single_child(display_filter_flow, "flow", fe, handle, path)
		return

	m = "flow filter {0} mode {1}".format(name, handle.mode)
	bcf = Tkinter.Frame(fe.stuff)
	if hasattr(handle, "divisor"):
		m += " divisor {0}".format(handle.divisor)
	if hasattr(handle, "baseclass"):
		bcmsg = " baseclass {0}".format(handle.baseclass)
		m += bcmsg
		temp = {"classid": handle.baseclass}
		btn = widget_classid(fe, temp, path[0][1], path[1][1], parent=bcf)

	if hasattr(handle, "baseclass"):
		Tkinter.Label(bcf, text=m).pack(fill=Tkinter.X, side=Tkinter.LEFT)
		btn.pack(side=Tkinter.RIGHT, fill=Tkinter.X)
	else:
		Tkinter.Label(bcf, text=m).pack(fill=Tkinter.X)
	bcf.pack(fill=Tkinter.X)

	pw = Tkinter.PanedWindow(fe.stuff)
	pw.columnconfigure(0, weight=1)
	pw.columnconfigure(1, weight=2)

	lf = Tkinter.Frame(pw)
	rf = Tkinter.Frame(pw)
	Tkinter.Label(lf, text="Extended matches:").pack(fill=Tkinter.X)
	Tkinter.Label(rf, text="Chain of actions:").pack(fill=Tkinter.X)

	lba = widget_actions(rf, handle.actions)
	lba.pack(fill=Tkinter.BOTH, expand=True)
	emt = widget_ematch(lf, " ".join(handle.ematch))
	emt.pack(fill=Tkinter.BOTH, expand=True)

	pw.add(lf), pw.add(rf)
	pw.pack(side=Tkinter.BOTTOM, fill=Tkinter.BOTH, expand=True)
	
	if handle.mode == 'hash':
		m = "Hash keys: {0}".format(handle.keys)
	elif handle.mode == 'map':
		m = "Map key: {0}".format(handle.keys)
		if hasattr(handle, 'map_rshift'):
			m += ", rshift {0}".format(handle.map_rshift)
		if hasattr(handle, 'map_addend'):
			m += ", addend 0x{0:x}".format(handle.map_addend)
		if hasattr(handle, 'map_or'):	m += ", or 0x{0:x}".format(handle.map_or)
		if hasattr(handle, 'map_xor'):	m += ", xor 0x{0:x}".format(handle.map_xor)
		if hasattr(handle, 'map_and'):	m += ", and 0x{0:x}".format(handle.map_and)
	Tkinter.Label(fe.stuff, text=m).pack(fill=Tkinter.X)

def display_filter_u32(fe, handle, path):
	[key, name] = path[-1]

	if len(handle.children) == 1:
		name = handle.fh.decode('ascii') if handle.fh else "u32"
		filter_single_child(display_filter_u32, name, fe, handle, path)
		return

	if not handle.fh:
		m = "u32 filter protocol {0} preference {1}"
		m = m.format(handle.proto, handle.pref)
		Tkinter.Label(fe.stuff, text=m).pack()
		Tkinter.Label(fe.stuff, text="Choose parts of the filter from the tree").pack(fill=Tkinter.BOTH, expand=True)
		return

	m = "u32 filter {0:x}:{1:x}:{2:x}"
	m = m.format(handle.htid, handle.bkt, handle.key)
	if hasattr(handle, 'link'):
		m += " link {0}".format(handle.link)
	if hasattr(handle, 'off'):
		eat = "eat" if handle.eat else ""
		m += " offset {0} at {1} {2}".format(handle.off, handle.offoff, eat)
	if hasattr(handle, 'hashmask'):
		m += " hash mask {0} at {1}".format(handle.hashmask, handle.hashoff)
	Tkinter.Label(fe.stuff, text=m).pack(fill=Tkinter.X)

	## Display button that goes to classid
	w = widget_classid(fe, handle, path[0][1], path[1][1])
	w.pack(fill=Tkinter.X)

	pw = Tkinter.PanedWindow(fe.stuff)
	pw.columnconfigure(0, weight=1)
	pw.columnconfigure(1, weight=2)

	lf = Tkinter.Frame(pw)
	rf = Tkinter.Frame(pw)
	Tkinter.Label(lf, text="u32 matches:").pack(fill=Tkinter.X)
	Tkinter.Label(rf, text="Chain of actions:").pack(fill=Tkinter.X)

	lba = widget_actions(rf, handle.actions)
	lba.pack(fill=Tkinter.BOTH, expand=True)

	pw.add(lf), pw.add(rf)
	pw.pack(side=Tkinter.BOTTOM, fill=Tkinter.BOTH, expand=True)

	matches = dict()
	matches_nh = dict()
	nhmask = (1 << 16) - 1
	for (key, mask, off) in sorted(handle.matches, key=lambda x: x[2]):
		x = "{0:08x}/{1:08x}".format(key, mask)
		where = matches_nh if off > nhmask else matches
		off = off & nhmask
		if off in where:
			where[off] = where[off] + "\n" + x
		else:
			where[off] = x
	if len(matches) == 0 and len(matches_nh) == 0:
		Tkinter.Label(lf, text="No matches!").pack(fill=Tkinter.BOTH, expand=True)
		return

	lbm = Tkinter.Listbox(lf, font="Courier")
	lbm.pack(fill=Tkinter.BOTH, expand=True)

	for i in range(0, max(matches)+1, 4):
		if i in matches:
			m = "off {0:2}: {1}".format(i, matches[i])
		else:
			m = "off {0:2}:".format(i)
		lbm.insert(Tkinter.END, m)
	if len(matches_nh) == 0:
		return
	for i in range(0, max(matches_nh)+1, 4):
		if i in matches_nh:
			m = "nh +{0:2}: {1}".format(i, matches_nh[i])
		else:
			m = "nh +{0:2}:".format(i)
		lbm.insert(Tkinter.END, m)


# ---------------------------------------- 8>< ===================
#  attach the functions, so the frontend can see them

from frontends.tkinter import tkinter_type2fn
tkinter_type2fn[ "filter" ] = display_filter
tkinter_type2fn[ "filter basic" ] = display_filter_basic
tkinter_type2fn[ "filter bpf" ] = display_filter_bpf
tkinter_type2fn[ "filter flow" ] = display_filter_flow
tkinter_type2fn[ "filter u32" ] = display_filter_u32

tkinter_type2fn[ "qdisc" ] = display_qdisc

