import paramiko, base64, hashlib

from collections import deque
from os import errno, stat
from socket import socket, getaddrinfo, gaierror, AF_UNSPEC, SOCK_STREAM, error as sockerror, timeout as socktimeo
from time import time, sleep
import json

from parsers.tc import TcFilter, TcParseLinks, TcParseQdiCls, TcParseFilters
from software.Linux import Linux
from software.OpenBSD import OpenBSD

def just_print(x):
	print(x)

def windoze_wait_hack():
	from sys import platform
	if platform == 'win32':
		from random import random
		sleep(random() / 10 + .01)

class Plan:
	"""One work unit scheduled to @next and then every @secs seconds."""
	def __init__(self, secs, fn):
		self.fn = fn
		self.change(secs)

	def change(self, secs):
		self.secs = secs
		self.next = time()

	def to_JSON(self):
		return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)

class Host(object):
	def __init__(self, name, timeout=10):
		self.curtask = None
		self.task_out = self.task_err = b''
		self.locked = self.connected = False

		self.unregister_all_software()

		self.name = name
		self.log = just_print
		self.timeout = timeout
		self.widget = self.sock = self.client = None
		self.gaipos = 0
		self.badness = self.went_bad = 0
		self.started = time()
		self.tasks = deque()
	
	def resolve_hostname(self, hostname, port=22):
		self.hostname = hostname
		self.port = port
		self.set_status("resolving {0} with port {1}".format(hostname, port))
		try:
			self.gai = getaddrinfo(hostname, port, AF_UNSPEC, SOCK_STREAM)
		except gaierror as e:
			self.gai = []
			self.set_status("gai error: {0}: {1}".format(hostname, e.strerror))
			return False
		return True

	def set_username_password(self, username, password):
		self.username = username
		self.password = password

	def __str__(self):
		return self.name
	
	def set_status(self, msg):
		self.status = msg
		self.log(self.name + ": " + str(msg))

	def set_logging_function(self, logfn):
		self.log = logfn

	def set_password_dialog(self, pwdlg):
		self.ask_for_pw = pwdlg

	def waiting_too_long(self):
		return time() - self.started > self.timeout

	def pop_task(self):
		"""POPing a task to schedule means I'm done with the previous one."""
		self.task_out = b''
		while self.tasks_in_queue():
			(cmd, cb, sudo, t) = self.tasks.popleft()
			if t and time() < t:
				self.tasks.append((cmd, cb, sudo, t))
				if self.tasks_in_queue() == 1:
					return None
				continue
			return (cmd, cb, sudo)
		return None

	def _that_bad(self):
		if self.badness > 30:
			howlong = time() - self.went_bad
			if howlong > 50:
				m = "{0} was unresponsive for {1:.1f} s, trying to reconnect."
				self.log(m.format(self.name, howlong))
				nw = map(lambda x: (x, self.plans[x].secs), self.plans.keys())
				nw += self.wishes
				self.disconnect()
				self.wishes = nw
				try:
					self.log("trying to connect...")
					self.connect()
					self.log("tried to connect.")
				except sockerror as se:
					self.log(se)
				return True
			m = "{0}: connection wasn't responding for {1:.1f} seconds now."
			self.log(m.format(self.name, howlong))
			return True
		return False
			
	def push_task(self, cmd, cb, use_sudo=False, when=None):
		if not self._that_bad():
			self.tasks.append((cmd, cb, use_sudo, when))
			return True
		return False

	def push_task_front(self, cmd, cb, use_sudo=False, when=None):
		if not self._that_bad():
			self.tasks.appendleft((cmd, cb, use_sudo, when))
			return True
		return False
	
	def tasks_in_queue(self):
		return len(self.tasks)

	def register_software(self, nsoft):
		"""Registers @nsoft with the current host.

		If @nsoft offers capabilities already covered, nothing happens."""

		stypes = nsoft.subtrees()
		ntypes = nsoft.node_types()
		for t in ntypes:
			if t in self.node_types:
				return False
		for (t, fn) in stypes:
			if t in self.subtrees:
				return False
		for t in ntypes:
			self.node_types[t] = nsoft
		for (t, fn) in stypes:
			self.subtrees[t] = fn
		self.software += [nsoft]
		return True

	def unregister_all_software(self):
		self.node_types = dict()
		self.subtrees = dict()
		# for things like refresh(all)
		self.software = []

		# Each subtree type can be checked periodically.
		# Some new tasks may potentially appear when these
		# things go off.  The tasks are then scheduled as usual.
		self.plans = dict()
		self.wishes = []

	def periodic_events(self, frontend):
		"""Executes any planned "periodic" events in our plan list."""
		nw = []
		for (t, secs) in self.wishes:
			if t in self.subtrees:
				self.plans[t] = Plan(secs, self.subtrees[t])
			else:
				nw += [(t, secs)]
		self.wishes = nw
		now = time()
		for t in self.plans:
			plan = self.plans[t]
			if now < plan.next:
				continue
			plan.next += plan.secs
			if not plan.fn(frontend):
				return False
		return True

	def set_plan(self, subtree, secs):
		"""Resets the periodic events for @subtree and makes a new plan."""
		if subtree in self.plans:
			self.plans[subtree].change(secs)
		elif subtree in self.subtrees:
			self.plans[subtree] = Plan(secs, self.subtrees[subtree])
		else:
			self.wishes += [(subtree, secs)]

	def unset_plan(self, subtree):
		"""If the user changed her mind about planning @subtree.

		Returns True if something was unset, False if not found."""
		nw = []
		ret = False
		for (t, s) in self.wishes:
			if t != subtree:
				nw += [(t, s)]
			else:
				ret = True
		self.wishes = nw
		if subtree in self.plans:
			del self.plans[subtree]
			return True
		return ret


class SSH_Host(Host):
	sudo_pwprompt = b"sudo(8) password prompt: "

	def __init__(self, name, timeout=10):
		super(SSH_Host, self).__init__(name, timeout=timeout)
		self.pkey = None
		self.session = None
		self.sudo_maybepw = False
		self.doing_upload = False

	def load_host_keys(self, file):
		"""Loads OpenSSH known_hosts file into paramiko's object."""
		if not self.client:
			self.client = paramiko.SSHClient()
		self.client.load_host_keys(file)

	def expected_key(self, cipher):
		"""Based on loaded host public keys, what should ours be?"""
		if not self.client:
			return None
		keys = self.client.get_host_keys()
		if self.hostname not in keys:
			return None
		if cipher not in keys[self.hostname]:
			return None
		return keys[self.hostname][cipher]

	def set_pubkey_policy(self, pol):
		if not self.client:
			self.client = paramiko.SSHClient()
		self.client.set_missing_host_key_policy(pol)

	def set_pubkey(self, b64string, cipher):
		key = paramiko.RSAKey(data=base64.decodestring(b64string))
		if not self.client:
			self.client = paramiko.SSHClient()
		self.client.get_host_keys().add(self.hostname, cipher, key)

	def set_identityfile(self, file, password=None):
		self.identityfile = file
		self.pkey = paramiko.RSAKey.from_private_key_file(file, password)

	def next_addr(self):
		st = 0
		if self.gaipos >= len(self.gai):
			return None
		while st != SOCK_STREAM and self.gaipos < len(self.gai):
			(f, st, p, name, addr) = self.gai[self.gaipos]
			self.gaipos = self.gaipos + 1
		return (f, st, addr)

	def connect(self, timeout=1000):
		if not self.gai:
			self.resolve_hostname(self.hostname, self.port)
		if not self.sock:
			next = self.next_addr()
			if not next:
				if len(self.gai) != 0:
					m = "ran out of {0} addresses, retrying"
					self.set_status(m.format(self.hostname))
					self.gaipos = 0
				return None
			(f, st, self.addr) = next

			self.sock = socket(f, st)
			self.set_status("connecting to {0}".format(self.addr))
			self.sock.setblocking(0)
			self.started = time()

		if not self.connected:
			try:
				windoze_wait_hack()
				self.sock.connect(self.addr)
			except sockerror as e:
				too_long = self.waiting_too_long()
				m = "waited too long" if too_long else e.strerror

				waiting = [ errno.EINPROGRESS, errno.EALREADY,
				    errno.EWOULDBLOCK ]
				if e.errno == errno.EISCONN:
					pass
				elif e.errno not in waiting or too_long:
					a = self.addr
					self.log("ERROR: {0}: {1}".format(a, m))
					self.sock.close()
					self.sock = None
					return self.connect(timeout)
				else:
					return self.sock.fileno()

		a = self.addr
		self.set_status("establishing SSH session with {0}".format(a))

		# At this point we've done our absolute best not to block the user.
		# But, if there is a SYN proxy in the way, it might have completed
		# the TCP handshake for us and we're gonna hang paramiko.SSHClient
		# anyway.  The 'timeout' kwarg doesn't seem to do anything useful.

		try:
			if self.pkey: 
				self.log("Decrypting the key can take up to 10 seconds!",\
				    setbusy=True)
				self.client.connect(self.hostname,	\
				    username=self.username,		\
				    pkey=self.pkey, sock=self.sock)
				self.log(None, setready=True)
			else:
				if not self.password:
					m = "You didn't enter password for {0}@{1}.  Do it now."
					m = m.format(self.username, self.hostname)
					self.password = self.ask_for_pw(m)
				self.log("{0}: using a password...".format(self),
				    setbusy=True)
				self.client.connect(self.hostname,	\
				    username=self.username, 		\
				    password=self.password, sock=self.sock)
				self.log(None, setready=True)
		except EOFError:
			self.set_status("{0} has closed connection".format(a))
			self.sock.close()
			self.sock = None
			self.log(None, setready=True)
			return None
		except paramiko.SSHException as se:
			self.log(se, setready=True)
			raise
		self.set_status("connection with {0} established".format(a))
		self.schroedinger = self.client.get_transport().open_session()
		self.schroedinger.settimeout(0.1)
		self.schroedinger.exec_command("cat")
		self.schroedinger_sent = False
		self.alive = ""
		self.connected = True

		# This will effectively wake it up.
		self.push_task("echo $PS1 && uname", self.os_based_on_uname)
		return self.sock.fileno()

	def disconnect(self):
		"""Close all sessions and prepare for reconnect."""
		self.set_status("disconnecting")
		self.connected = False
		self.unregister_all_software()
		if self.session:
			try:	self.session.close()
			# the session may try to send the buffer, and fail
			except EOFError:
				pass
		if self.sock:
			self.sock.close()
		if hasattr(self, 'schroedinger'):
			try:	self.schroedinger.close()
			except:	pass
			finally:
				self.schroedinger = None
				self.schroedinger_sent = False
		self.session = self.sock = self.gai = None
		self.gaipos = self.badness = 0
		self.set_status("gone now.")

	def check_if_exited(self):
		self.check_sudo()

		if self.session.exit_status_ready():
			# read the rest of the buffer
			try:
				ndata = self.session.recv(65536)
				while len(ndata):
					self.task_out = self.task_out + ndata
					ndata = self.session.recv(65536)
			except socktimeo:
				return False

			# finish this session
			self.errcode = self.session.recv_exit_status()
			self.session.close()
			self.session = None
			(self.lastcmd, cb, _) = self.curtask
			self.curtask = None
			if self.errcode:
				m = "{0}: '{1}' exited with code {2}"
				self.log(m.format(self.name, self.lastcmd, self.errcode))
			#else:
			#	m = "{0}: '{1}' exited normally"
			#	self.log(m.format(self.name, self.lastcmd))
			cb(self.task_out)
			return True
		return False

	def check_sudo(self):
		"""Returns True if sudo(8) has been dealt with already."""
		if self.sudo_maybepw and self.session.recv_ready():
			# recv on stdout is ready, presuming no sudo pw!
			self.sudo_maybepw = False
		elif self.sudo_maybepw and self.session.exit_status_ready():
			# something went either horribly wrong or quick
			self.errcode = self.session.recv_exit_status()
			if self.session.recv_stderr_ready():
				self.log(self.session.recv_stderr(256))
			else:	# log the last message, probably from sudo
				a = self.sudo_err.split(b'\n')
				for l in reversed(a):
					if not l:
						continue
					self.log(l)
					break
			self.sudo_maybepw = False
		elif self.sudo_maybepw:
			try:
				self.sudo_err += self.session.recv_stderr(256)
			except socktimeo:
				return False
			while self.sudo_err.endswith(self.sudo_pwprompt):
				passwd = self.ask_for_pw(self.sudo_err)
				self.sudo_err = b''
				if passwd:
					self.session.send(passwd + '\n')
				elif hasattr(self, 'password') and self.password:
					self.session.send(self.password + '\n')
				else:
					self.log("no sudo password")
				return False
		return not self.sudo_maybepw

	def sudo(self, command):
		assert not self.session
		self.errcode = None
		t = self.client.get_transport()
		try:	self.session = t.open_session()
		except:
			self.set_status("can't open_session, disconnecting...")
			self.disconnect()
			return None
		self.session.setblocking(0)
		self.sudo_err = b''
		self.sudo_maybepw = True

		# If we do session.get_pty(), stdout and stderr will become one.
		# We can't have that under any circumstances.
		self.session.set_combine_stderr(False)

		windoze_wait_hack()

		# Security rant:  reading the password from stdin might
		# seem wrong, but the only way to intercept stdin is to
		# change your default login shell that executes sudo(8)
		# and in that case your system has had it anyway. 

		cmd = "/usr/bin/sudo -p '{0}' -S {1}"
		cmd = cmd.format(self.sudo_pwprompt.decode('ascii'), command)
		try:
			self.session.exec_command(cmd)
		except EOFError:
			self.set_status("{0} closed connection".format(self.name))
			self.disconnect()
			return None

		# We won't do self.check_if_exited here, it might be too soon.
		# select() will wake us up when it's time.
		return self.session

	def cat_alive(self):
		if not self.schroedinger_sent:
			self.schroedinger_sent = True
			try:
				self.schroedinger.send("alive\n")
			except EOFError:
				self.set_status("cat at {0} closed connection".format(self.name))
				self.disconnect()
				return False
			self.alive = ""
		if self.alive != 'alive\n':
			try:
				self.alive += self.schroedinger.recv(1024).decode('ascii')
			except socktimeo as st:
				pass
			except sockerror as se:
				self.log(self.name + " cat read: " + str(se))
			if self.alive != 'alive\n':
				if self.badness in [5, 15, 35]:
					m = "{0}: connection not responding for {1:.1f} seconds"
					self.log(m.format(self.name, time() - self.went_bad))
				elif self.badness == 0:
					self.went_bad = time()
				self.badness += 1
				return False
		self.schroedinger_sent = False
		self.badness = 0
		self.went_bad = 0
		return True
				
	def run(self, command):
		### XXX logging NOTICE nebo DEBUG
		self.set_status("executing command '{0}'".format(command))
		if self.session:
			from traceback import print_stack
			print_stack()
			return
		self.errcode = None

		t = self.client.get_transport()
		try:	self.session = t.open_session()
		except:
			self.set_status("can't open_session, disconnecting...")
			self.disconnect()
			return None
		self.session.setblocking(0)
		windoze_wait_hack()
		try:
			self.session.exec_command(command)
		except EOFError:
			self.set_status("{0} closed connection".format(self.name))
			self.disconnect()
			return None
		self.check_if_exited()
		return self.session

	def move_forward(self, can_write=False):
		"""Tries to progress the communication further, True if done."""
		if not self.check_sudo():
			return False
		if self.doing_upload and can_write:
			# XXX this is still unfinished, because SSH has its own
			# window mechanism, controlled by the other thread
			# It seems we don't have a reliable multiplexer over
			# paramiko's _ready() events.
			# XXX this needs lot more testing.

			self.log("started with {0}".format(self.sent))
			if len(self.waiting) != 0:
				self.log("first "+str(self.session.send_ready()))
				try:
					sent = self.session.send(self.waiting)
				except socktimeo:
					self.log("1st timeo")
					return False
				self.waiting = self.waiting[sent:]
				self.sent += sent
				self.log("1st + "+str(sent))
			while self.session.send_ready():
				if len(self.waiting) == 0:
					self.waiting = self.scp_src.read(65535)
					self.scp_md5.update(self.waiting)
				if len(self.waiting) == 0:
					self.log("ending it")
					try:
						self.scp_put_finish()
					except Exception as e:
						self.log(e)
					self.scp_src.close()
					del self.scp_src
					took = time() - self.scp_started
					self.log("transfer took {0} seconds".format(took))
					break
				else:
					try:
						sent = self.session.send(self.waiting)
					except socktimeo:
						self.log("timeo")
						break
					self.log("+ "+str(sent))
					self.waiting = self.waiting[sent:]
					self.sent += sent
			self.log("sent {0}".format(self.sent))
			return False
		else:
			try:
				ndata = self.session.recv(65536)
			except socktimeo:
				ndata = ""
			if len(ndata):
				self.task_out = self.task_out + ndata
		return self.check_if_exited()

	def scp_put(self, local, remote):
		if "'" in remote:
			return
		t = self.client.get_transport()
		self.session = t.open_session()
		self.doing_upload = True
		self.session.exec_command("scp -t '{0}'".format(remote))
		response = self.session.recv(512)
		if not response:
			raise Exception("EOF response scp put")
		if response[0] != '\x00':
			raise Exception("!0 response scp put "+response)

		sb = stat(local)
		mode = oct(sb.st_mode)[-4:]
		name = remote.split('/')[-1]
		scp_cmd = "C{0} {1} {2}\n".format(mode, sb.st_size, name)
		self.log(scp_cmd)
		self.session.sendall(scp_cmd)

		response = self.session.recv(512)
		if not response:
			raise Exception("EOF response scp put C")
		if response[0] != '\x00':
			raise Exception("!0 response scp put C "+response)

		# the session is now ready to transfer data
		self.scp_src = open(local)
		self.scp_md5 = hashlib.md5()
		self.scp_filename = remote
		self.sent = 0
		self.waiting = self.scp_src.read(65535)
		self.scp_md5.update(self.waiting)
		self.scp_started = time()
		self.session.setblocking(0)

	def scp_put_finish(self):
		# this will just be short :-)
		self.session.setblocking(1)
		self.session.sendall('\x00')
		
		response = self.session.recv(512)
		self.session.close()
		self.session = None
		self.doing_upload = False

		if not response:
			raise Exception("EOF response scp put fin")
		if response[0] != '\x00':
			raise Exception("!0 response scp put fin "+response)

	def whatis(self, path):
		"""Returns the type and handle of the thing described by @path."""

		lasttype = path[-1][0]
		for sw in self.software:
			attempt = sw.whatis(path)
			if attempt[0][0] == lasttype:
				return attempt
		# At the beginning, we might not have any software registered.
		return (["host"], self)

	def recv_ready(self):
		return self.session.recv_ready() if self.session else False
	def send_ready(self):
		return self.session.send_ready() if self.session else False

	def os_based_on_uname(self, o, e=None):
		if e:
			self.log("UNAME FAILED: " + e)
		a = o.split(b'\n')
		self.ps1 = a[0]
		u = a[1].strip()
		if u == b"Linux":
			sw = Linux()
			self.push_task("ls /etc/", sw.linux_based_on_etc)
		elif u == b"OpenBSD":
			sw = OpenBSD()
			self.push_task("uname -r", sw.openbsd_based_on_uname_r)
		else:
			self.log("unknown system {0}".format(u))
			return
		if not self.register_software(sw):
			self.log("{0}: can't register {1}".format(self.name, sw))
			self.log(self.ps1)
			return
		sw.host = self
		self.log("{0}: registered {1}".format(self.name, sw))


def ssh_key_digest(k):
	"""Given a paramiko.RSAKey, returns string(MD5 digest) like OpenSSH."""
	if not k:
		return "no key"
	digest = hashlib.md5(k.asbytes()).hexdigest()
	d = ""
	for i in range(0, len(digest), 2):
		d = d + "{0}{1}:".format(digest[i], digest[i+1])
	return d[:-1]

