from software import RootPriviledgesNeededException

def bandwidth(s):
	s = s.strip()
	deduct = 4 if s.endswith('it') else 2		# Kb vs. Kbit
	if s.endswith('Tbit') or s.endswith('Tb'):
		f = float(s[:-deduct]) * 1000 * 1000 * 1000 * 1000
	elif s.endswith('Gbit') or s.endswith('Gb'):
		f = float(s[:-deduct]) * 1000 * 1000 * 1000
	elif s.endswith('Mbit') or s.endswith('Mb'):
		f = float(s[:-deduct]) * 1000 * 1000
	elif s.endswith('Kbit') or s.endswith('Kb'):
		f = float(s[:-deduct]) * 1000
	else: # 37b
		f = float(s[:-(deduct-1)])
	return (f)

def datasize(s):
	s = s.strip()
	if s.endswith('b'):
		f = int(s[:-1])
	return (f)

def hfsc_d(s):
	s = s.strip()
	if s.endswith('us'):
		f = float(s[:-2]) * 1000 * 1000
	elif s.endswith('ms'):
		f = float(s[:-2]) * 1000
	elif s.endswith('s'):
		f = float(s[:-1])
	return int(f)

# Encode/Decode major/minor numbers off a qdisc/class handle
def majmin_encode(major, minor, third=None):
	if third is not None:
		return '{0:x}:{1:x}:{2:x}'.format(major, minor, third)
	return '{0:x}:{1:x}'.format(major, minor)

def majmin_decode(data):
	return map(lambda x: int(x, 16) if x else 0, data.split(b':'))


def TcParseHFSC(s):
	"""Parses HFSC parameters and returns tuples like from ALTQ."""
	mapping = {b'rt': 'realtime', b'ls': 'linkshare', b'ul': 'upperlimit'}
	a = s.strip().split()
	curves = dict()
	for i in range(0, len(a)):
		if a[i] in [b'rt', b'ls', b'ul', b'sc']:
			if (a[i+1], a[i+3], a[i+5]) != (b'm1', b'd', b'm2'):
				m = "hfsc tc weird line >>{0}<<".format(line)
				raise Exception(m)
			(m1, d, m2) = (a[i+2], a[i+4], a[i+6])
			(m1, d, m2) = map(lambda x: x.decode('ascii'), (m1, d, m2))
			(m1, d, m2) = (bandwidth(m1), hfsc_d(d), bandwidth(m2))
			if a[i] == b'sc':
				curves['realtime'] = (m1, d, m2)
				curves['linkshare'] = (m1, d, m2)
			else:
				curves[mapping[a[i]]] = (m1, d, m2)
	return curves

def TcParseHTB(s):
	"""Parses HTB parameters."""
	a = s.strip().split()
	param = dict()
	for i in range(0, len(a)):
		k = a[i].decode('ascii')
		if k == 'prio':
			param[k] = int(a[i+1])
		elif k in ['rate', 'ceil']:
			param[k] = bandwidth(a[i+1].decode('ascii'))
		elif k in ['burst', 'cburst']:
			param[k] = datasize(a[i+1].decode('ascii'))
	return param

interface_forbidden_keys = ['ifindex', 'ifname', 'flags', 'egress', 'ingress', 'lladdr']
def TcParseLinks(s):
	ifs = dict()
	if_lines = []
	for line in s.split(b'\n'):
		if not line:
			continue
		elif line[0] != ord(' '):
			iface = dict()
			f = line.split()
			if len(f) < 4:
				raise Exception("short line: {0}".format(line))
			# colons at the end of these are bad
			iface['ifindex'] = int(f[0].strip()[:-1])
			iface['ifname'] = f[1][:-1].decode('ascii')
			iface['flags'] = f[2].decode('ascii')
			iface['egress'] = dict()
			iface['ingress'] = dict()

			f = f[3:]
			while len(f) > 1:
				key = f[0].decode('ascii')
				if key in ['mtu', 'qlen']:
					iface[key] = int(f[1])
				elif key not in interface_forbidden_keys:
					iface[key] = f[1]
				f = f[2:]
			ifs[iface['ifname']] = iface
		else:
			s = line.strip().split()
			if s[0] == b'link/ether':
				iface['lladdr'] = s[1].decode('ascii')
			#print(iface['ifname'] + line)
	#print(s)
	return ifs


def TcParseQdiClsRootParent(what, f, one, major):
	"""Parses qdisc's or class's "root"/"parent" options."""
	if f[0] == b'root':
		# BAD_BAD_TC telling us wrong 'root' classes
		# they are actually parented by <MAJOR>:0
		# which is either a class or a classful qdisc
		if what == b'class':
			one['parent'] = majmin_encode(major, 0)
		one['rest'] = b' '.join(f[1:])
	elif f[0] == b'parent':
		one['parent'] = majmin_encode(*majmin_decode(f[1]))
		one['rest'] = b' '.join(f[2:])
	else:
		# centos 5 won't tell us "root" at the root level
		one['rest'] = b' '.join(f)

def TcParseQdiCls(s, ifname=None):
	"""Parses the output from "tc qdisc show" or "tc class show"."""
	allifs = dict()
	one = None
	if ifname:
		all = dict()
	for line in s.split(b'\n'):
		if not line:
			continue
		elif line.startswith(b'qdisc ') or line.startswith(b'class '):
			one = dict()
			one['qdiscs'] = dict()
			one['classes'] = dict()
			one['filters'] = dict()
			f = line.split()
			if len(f) < 4:
				raise Exception("short line: {0}".format(line))
			one['qdcl'] = f[0].decode('ascii')
			one['type'] = f[1].decode('ascii')

			(major, minor) = majmin_decode(f[2])
			one['handle'] = majmin_encode(major, minor)

			if f[1] == b'ingress':
				# "qdisc ingress ffff: dev enp1s0f1 THE_REST"
				thisiface = f[4].decode('ascii')
				# we don't want any bonkers parents here
				one['rest'] = b' '.join(f[5:])
				one['ifname'] = thisiface
				if ifname:
					all[f[2]] = one
				else:
					if not thisiface in allifs:
						allifs[thisiface] = dict()
					allifs[thisiface][one['handle']] = one

			elif ifname and f[3] in [b'root', b'parent']:
				TcParseQdiClsRootParent(f[0], f[3:], one, major)
				all[one['handle']] = one

			elif not ifname and f[3] == b'dev':
				# "qdisc pfifo_fast 0: dev enp3s0 THE_REST"
				thisiface = f[4].decode('ascii')
				one['ifname'] = thisiface
				if thisiface not in allifs:
					allifs[thisiface] = dict()
				TcParseQdiClsRootParent(f[0], f[5:], one, major)

				# BAD_BAD_TC	This is REALLY stupid.  A handle
				# on multiple levels of the tree represents many
				# different nodes (pfifo_fast's under mq).
				# Because for us the root is kind of important,
				# we're gonna assume the ordering of "tc qdisc"
				# output and relabel any subordinates.
				if one['handle'] in allifs[thisiface]:
					newhandle = one['parent']+'->'+one['handle']
					allifs[thisiface][newhandle] = one
				else:
					allifs[thisiface][one['handle']] = one

			else:
				raise Exception("bad line: {0}".format(line))

		else:
			if not one:
				raise Exception("weird line: {0}".format(line))
			one['rest'] += line
	return allifs if not ifname else all

def TcParseBPF(line):
	from parsers.bpf import BPF
	ret = []
	targets = {}
	rest = line.split(b',')
	count, total = 0, int(rest[0])
	for l in rest[1:]:
		code = tuple(map(int, l.split()))
		disas, branch = BPF.dis(code)
		if branch:
			branch += count + 1
			targets[branch] = True
			disas += ', L{0}'.format(branch)

		if count in targets:
			ret += ['L{0}:\t{1}\t{2}'.format(count, disas, code)]
		else:
			ret += ['{0}\t{1}'.format(disas, code)]
		count += 1
	assert(count == total)
	return ret

def TcParseEmatch(line):
	allowed_ematches = ["cmp", "meta", "nbyte", "ipset", "u32"]
	def parse_expr(line):
		(node, rest) = parse_term(line)
		cur = dict()
		if rest.startswith(b'AND'):
			cur['type'] = "and"
		elif rest.startswith(b'OR'):
			cur['type'] = "or"
		elif not rest or rest[0] == ord(')'):
			return (node, rest)
		cur['left'] = node
		(cur['right'], rest) = parse_expr(rest[3:].lstrip())
		return (cur, rest)
	def parse_term(line):
		cur = dict()
		if line.startswith(b'[unknown ematch'):
			nextem = line.find(b']')+1
			current = line[:nextem]
			cur['type'] = "unknown"
			cur['data'] = current
			return (cur, line[nextem:].lstrip())
		inverted = False
		if line.startswith(b'NOT'):
			inverted = True
			line = line[3:].lstrip()
		if line[0] == ord('('):
			(cur, line) = parse_expr(line[1:].lstrip())
			if line[0] != ord(')'):
				raise Exception("unbalanced parens {0}".format(line))
			if inverted:
				cur['inverted'] = True
			return (cur, line[1:].lstrip())
		if inverted:
			cur['inverted'] = True
		return (cur, parse_match(line, cur))
	def parse_match(line, node):
		a = line.split(b'(')
		module = a[0].strip().decode('ascii')
		if module not in allowed_ematches:
			raise Exception("unknown ematch type " + module)

		if line.strip().endswith(b'ipset('):
			raise RootPriviledgesNeededException("ipset ematch")

		# consume the left paren as well...
		line = line.strip()[len(module)+1:].lstrip()
		if module == 'cmp':	r = parse_cmp(line, node)
		elif module == 'meta':	r = parse_meta(line, node)
		elif module == 'nbyte':	r = parse_nbyte(line, node)
		elif module == 'ipset':	r = parse_ipset(line, node)
		elif module == 'u32':	r = parse_u32(line, node)
		else:
			raise Exception("unknown2 ematch type " + module)
		return r[1:].lstrip()
	def parse_cmp(line, node):
		a = line.split()
		(align, at, offset) = (a[0], a[1], int(a[2]))
		if align not in [b'u8', b'u16', b'u32'] or at != b'at':
			raise Exception("cmp ematch wrong >>{0}<<".format(line))
		off = 3
		layer = mask = trans = False
		if a[off] == b'layer':
			layer = int(a[off+1])
			off += 2
		if a[off] == b'mask':
			mask = int(a[off+1], 16)
			off += 2
		if a[off] == b'trans':
			trans = True
			off += 1
		cmp = a[off]
		if cmp not in [b"eq", b"lt", b"gt"]:
			raise Exception("cmp ematch wrong cmp >>{0}<<".format(cmp))
		value = a[off+1].decode('ascii')
		if value[-1] == ')':
			value = value[:-1]
			ret = b')'
		else:
			ret = b''

		(align, cmp) = map(lambda x: x.decode('ascii'), (align, cmp))
		keys = ['align', 'layer', 'mask', 'offset', 'op', 'trans', 'type', 'value']
		vals = [align, layer, mask, offset, cmp, trans, 'cmp', int(value)]
		for k, v in zip(keys, vals):
			node[k] = v

		return ret + b' '.join(a[off+2:])
	def parse_meta(line, node):
		def parse_meta_id(a):
			lval_id = a[0].decode('ascii')
			off = 1
			shift = mask = 0
			if a[off] == b'shift':
				shift = int(a[off+1])
				off += 2
			if a[off] == b'mask':
				mask = int(a[off+1], 16)
				off += 2
			return ((lval_id, shift, mask), a[off:])
			
		end = line.find(b')')
		a = line[:end].strip().split()
		if a[0].isdigit():
			lval = int(a[0])
			op = a[1].decode('ascii')
			off = 2
		else:
			(lval, a) = parse_meta_id(a)
			op = a[0].decode('ascii')
			off = 1

		rval = a[off]
		if rval.isdigit():
			rval = int(rval)
		else:
			(rval, a) = parse_meta_id([rval] + a[off+1:])
		(node['lval'], node['rval'], node['op']) = (lval, rval, op)
		node['type'] = 'meta'
		return line[end:]
	def parse_ipset(line, node):
		# XXX This has to be run as root!  Otherwise things go VERY wrong!
		end = line.find(b')')
		a = line[:end].strip().split()
		(node['setname'], node['flags']) = (a[0], a[1])
		node['type'] = 'ipset'
		return line[end:]
	def parse_u32(line, node):
		end = line.find(b')')
		a = line[:end].strip().split()
		(valuemask, at) = (a[0], a[1])
		if at != b"at":
			raise Exception("u32 ematch wrong >>{0}<<".format(line))
		if a[2].startswith(b'nexthdr+'):
			offset = int(a[2].split(b'+')[1]) | (1 << 16)
		else:
			offset = int(a[2])
		(value, mask) = map(lambda x: int(x, 16), valuemask.split(b'/'))
		(node['value'], node['mask'], node['offset']) = (value, mask, offset)
		node['type'] = 'u32'
		return line[end:]
	def parse_nbyte(line, node):
		line = line.lstrip()
		endneedle = line.find(b'"')

		needlehex = line[:endneedle].strip()
		nchars = 0
		chars = []
		for hex in needlehex.split():
			chars += [int(hex, 16)]
			nchars += 1

		# just skip the confusing string representation in quotes
		line = line[endneedle+1:].lstrip()
		line = line[nchars+2:].lstrip()

		a = line.split()
		if a[0] != b'at':
			raise Exception("nbyte wrong line >>{0}<<".format(line))
		offset = a[1]
		ret = b''
		layer = '1'
		if a[2] == b'layer':
			layer = a[3]
			if layer[-1] == ord(')'):
				layer = layer[:-1]
				ret = b')'
			off = 4
		else:
			if offset[-1] == ord(')'):
				offset = offset[:-1]
				ret = b')'
			off = 2
		(node['layer'], node['offset']) = map(int, (layer, offset))
		(node['needle'], node['type']) = (chars, 'nbyte')
		return ret + b' '.join(a[off:])

	# This is actually the code of TcParseEmatch,
	# the rest of it is just parsing functions.
	line = line.strip()
	if line:
		(tree, rest) = parse_expr(line)
		return (tree, rest)
	return None

def TcPrintEmatch(n):
	i = "not " if 'inverted' in n else ""
	if n['type'] in ['and', 'or']:
		return i+n['type']
	elif n['type'] == 'cmp':
		m = "cmp {0} at {1} {2} {3}"
		return i+m.format(n['align'], n['offset'], n['op'], n['value'])
	elif n['type'] == 'ipset':
		m = "ipset {0} flags: {1}"
		return i+m.format(n['setname'], n['flags'])
	elif n['type'] == 'meta':
		m = "meta {0} {1} {2}"
		return i+m.format(n['lval'], n['op'], n['rval'])
	elif n['type'] == 'nbyte':
		m = "nbyte {0} at {1} layer {2}"
		return i+m.format(n['needle'], n['offset'], n['layer'])
	elif n['type'] == 'u32':
		(offset, nmask) = (int(n['offset']), (1 << 16) - 1)
		if offset > nmask:
			offset = "nexthdr+{0}".format(offset & nmask)
		m = "u32 {0:08x}/{1:08x} at {2}"
		return i+m.format(n['value'], n['mask'], offset)
	elif n['type'] == 'unknown':
		return i+n['data']
	# seriously, this has to be in sync with the parser...
	return i+"potazmo"

def TcNewAction(t, order=None):
	if t == b"csum":
		last = TcActionCsum(order)
	elif t == b"ipt":
		last = TcActionIpt(order)
	elif t == b"mirred":
		last = TcActionMirred(order)
	elif t == b"nat":
		last = TcActionNat(order)
	elif t == b"pedit":
		last = TcActionPedit(order)
	elif t == b"police":
		last = TcActionPolice(order)
	elif t == b"skbedit":
		last = TcActionSkbedit(order)
	else:
		last = TcActionGact(order)
		last.set_type(t)
	return last

class TcAction(object):
	state = "action"
	def __init__(self, order):
		if order:
			self.order = int(order.split(b":")[0])

	def __str__(self):
		fmt = "action {0} [{1}]"
		return fmt.format(self.type, self.order)

	def add_line(self, line):
		a = line.split()
		for i in range(0, len(a)):
			if a[i] == b'index':
				self.index = int(a[i+1])
			elif a[i] == b'bind':
				self.bindcnt = int(a[i+1])
			elif a[i] == b'ref':
				self.refcnt = int(a[i+1])

class TcActionCsum(TcAction):
	type = "csum"
	def parameters(self, a):
		reading_protos = reading_action = False
		self.protos = []
		for i in range(0, len(a)):
			if reading_protos:
				self.protos += [a[i][:-1].decode('ascii')] # comma or right paren
			if a[i][0] == ord('('):
				self.protos += [a[i][1:-1].decode('ascii')]
				reading_protos = True
			if a[i][-1] == ord(')'):
				reading_protos = False
			if reading_action:
				self.action = TcNewAction(a[i])
				self.action.parameters(a[i+1:])
				reading_action = False
				return
			if a[i] == b'action':
				reading_action = True

	def __str__(self):
		return "csum {0} action: {1}".format(self.protos, self.action)

class TcActionGact(TcAction):
	types = [b'reclassify', b'continue', b'drop', b'pass', b'ok', b'pipe', b'stolen']
	def parameters(self, a):
		for i in range(0, len(a)):
			if a[i] == b'action':
				self.type = a[i+1].decode('ascii')
			elif a[i] == b'overhead':
				self.overhead = a[i+1]

	def set_type(self, t):
		if t in self.types:
			self.type = t.decode('ascii')
		elif t != b'gact':
			raise Exception("bad action type {0}".format(t))

	def __str__(self):
		return self.type
		
class TcActionMirred(TcAction):
	type = "mirred"
	def parameters(self, a):
		for i in range(0, len(a)):
			if a[i][0] == ord('('): # opening paren
				self.direction = a[i][1:].decode('ascii').lower()
			elif a[i] == b'Mirror':
				self.type = 'mirror'
			elif a[i] == b'Redirect':
				self.type = 'redirect'
			elif a[i] == b'to' and a[i+1] == b'device':
				# strip out the closing paren
				self.dev = a[i+2][:-1].decode('ascii')
				self.action = TcNewAction(a[i+3])
				self.action.parameters(a[i+4:])

	def __str__(self):
		return "{0} {1} -> {2}".format(self.type, self.direction, self.dev)
		
class TcActionNat(TcAction):
	type = "nat"
	def parameters(self, a):
		self.direction = a[0].decode('ascii')
		self.oldipv4 = a[1].decode('ascii')
		self.newipv4 = a[2].decode('ascii')
		self.action = TcNewAction(a[3])
		self.action.parameters(a[3:])

	def __str__(self):
		m = "{0} {1} {2} -> {3} (action {4})"
		return m.format(self.type, self.direction, self.oldipv4, self.newipv4, self.action)
		
class TcActionPedit(TcAction):
	type = "pedit"
	def parameters(self, a):
		self.keys = []
		for i in range(0, len(a)):
			if a[i] == b'action':
				self.action = TcNewAction(a[i+1])
				# "keys" follow, so who knows? XXX
				#self.action.parameters(a[3:])
			elif a[i] == b'keys':
				self.nkeys = int(a[i+1])

	def __str__(self):
		m = "{0} keys: {1}, action: {2}"
		return m.format(self.type, self.nkeys, self.action)

	def add_line(self, l):
		f = l.strip().split()

		# --->	b'key #0 at -8: val ffffffff mask ffffffff'
		if len(f) > 6 and f[0] == b'key' and f[2] == b'at' and f[4] == b'val':
			(key, off, val) = (f[1][1:], f[3][:-1], f[5])
			mask = int(f[7], 16) if f[6] == b'mask' else 0
			(off, val) = (int(off), int(val, 16))

			if int(key) != len(self.keys):
				raise Exception("pedit, wrong key " + key)
			self.keys = self.keys + [(off, val, mask)]

		# --->  b' \t index 2 ref 1 bind 1'
		elif len(f) >= 6 and f[0] == b'index' and f[2] == b'ref' and \
		    f[4] == b'bind':
			self.index = int(f[1])
			(self.ref, self.bind) = (int(f[3]), int(f[5]))

		elif l.strip():
			raise Exception("pedit unknown line >>{0}<<".format(l))

class TcActionSkbedit(TcAction):
	type = "skbedit"
	def parameters(self, a):
		for i in range(0, len(a)):
			if a[i] == b'queue_mapping':
				self.queue_mapping = int(a[i+1])
			elif a[i] == b'mark':
				self.mark = int(a[i+1])
			elif a[i] == b'priority':
				self.priority = majmin_encode(*majmin_decode(a[i+1]))

			# skbedit output has a bug that omits one newline
			# therefore, on "Action statistics:" we change our state
			elif a[i] == b'Action' and a[i+1] == b'statistics:':
				state = "statistics"

	def __str__(self):
		qm = " queue_mapping: {0}".format(self.queue_mapping) if hasattr(self, 'queue_mapping') else ""
		mark = " mark: {0}".format(self.mark) if hasattr(self, 'mark') else ""
		priority = " priority: {0}".format(self.priority) if hasattr(self, 'priority') else ""
		return "skbedit{0}{1}{2}".format(qm, mark, priority)

class TcActionPolice(TcAction):
	type = "police"
	def parameters(self, a):
		self.handle = int(a[0], 16)
		a = a[1:]
		for i in range(0, len(a) - 1):
			val = a[i+1].decode('ascii')
			if a[i] == b'rate':
				self.rate = bandwidth(val)
			elif a[i] == b'burst':
				self.burst = bandwidth(val)
			elif a[i] == b'mtu':
				self.mtu = bandwidth(val)
			elif a[i] == b'action':
				self.action = TcNewAction(a[i+1])
				self.action.parameters(a[i+1:])

	def __str__(self):
		m = "{0} rate {1} burst {2} -> {3}"
		return m.format(self.type, self.rate, self.burst, self.action)


class TcFilter(object):
	def __init__(self, pref, proto, parent=None):
		self.pref = int(pref)
		self.proto = proto
		self.parent = parent
		self.children = dict()
		self.actions = self.matches = self.ematch = []
		self.widget = None

	def __str__(self):
		fmt = "filter {0} [{1}/{2}]"
		return fmt.format(self.type, self.proto, self.pref)

	def identify(self):
		"""By default, pref identifies us; not if you're a child."""
		return self.pref

	def update_frontend(self, fe, rootwidget):
		(t, i) = ("filter", self.identify())
		if self.widget:
			fe.ltv_update_node(self.widget, str(self), t, i)
		else:
			w = fe.ltv_add_node(rootwidget, str(self), t, i)
			self.widget = w

		for h in sorted(self.children):
			ch = self.children[h]
			ch.update_frontend(fe, self.widget)

	def delete_frontend(self, fe):
		fe.ltv_delete_node(self.widget)
		self.widget = None

	def probably_ematch(self, line):
		self.ematch = self.ematch + [line]

	def add_line(self, line):
		raise Exception("UNKNOWN LINE AT {0}\n{1}".format(self.type, line))

	# XXX wipe these out
	def add_action(self, a):
		self.actions = self.actions + [a]
	def add_match(self, a):
		self.matches = self.matches + [a]

class TcFilterBasic(TcFilter):
	type = "basic"
	def parameters(self, a):
		self.handle = None	# if it's the root filter node
		for i in range(0, len(a)):
			if a[i] == b'handle':
				self.handle = int(a[i+1], 16)
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))
		return self.handle

	def __str__(self):
		if not hasattr(self, 'handle') or not self.handle:
			return super(type(self), self).__str__()
		m = "handle 0x{0:x}".format(self.handle)
		if hasattr(self, "classid"):
			m = m + " classid {0}".format(self.classid)
		return m

	def add_filter(self, a):
		child = TcFilterBasic(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		self.children[handle] = child
		return child

	def add_line(self, l):
		self.probably_ematch(l)

	def identify(self):
		if hasattr(self, 'handle') and self.handle:
			return self.handle
		return self.pref

class TcFilterBPF(TcFilter):
	type = "bpf"
	def parameters(self, a):
		self.handle = None
		bytecode = False
		for i in range(0, len(a)):
			if a[i] == b'handle':
				self.handle = int(a[i+1], 16)
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))
			elif a[i] == b'bytecode':
				self.bytecode = b''
				bytecode = True
			elif bytecode:
				self.bytecode += b' '
				self.bytecode += a[i]
				if a[i][-1] == ord("'"):
					bytecode = False
					self.bytecode = self.bytecode.strip(b" '")
		assert(bytecode == False)
		return self.handle

	def __str__(self):
		if not hasattr(self, 'handle') or not self.handle:
			return super(type(self), self).__str__()
		m = "handle 0x{0:x}".format(self.handle)
		if hasattr(self, "classid"):
			m = m + " classid {0}".format(self.classid)
		if hasattr(self, "bytecode"):
			codelen = int(self.bytecode[:self.bytecode.find(b',')])
			m = m + " bytecode {0} insn's".format(codelen)
		return m

	def add_filter(self, a):
		child = TcFilterBPF(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		self.children[handle] = child
		return child

	def identify(self):
		if hasattr(self, 'handle') and self.handle:
			return self.handle
		return self.pref

class TcFilterFlow(TcFilter):
	type = "flow"
	mode = "none" # hash or map
	def parameters(self, a):
		self.handle = None	# if it's the root filter node
		for i in range(0, len(a)):
			if a[i] == b'handle':
				self.handle = int(a[i+1], 16)
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))

			elif a[i] == b'baseclass':
				self.baseclass = majmin_encode(*majmin_decode(a[i+1]))
			elif a[i] == b'divisor':
				self.divisor = int(a[i+1]) ## XXX

			elif a[i] == b'hash':
				self.mode = 'hash'
			elif a[i] == b'perturb':
				self.perturb = int(a[i+1][:-3])	## b'42sec'

			elif a[i] == b'map':
				self.mode = 'map'
			elif a[i] == b'and':
				self.map_and = int(a[i+1], 16)
			elif a[i] == b'or':
				self.map_or = int(a[i+1], 16)
			elif a[i] == b'xor':
				self.map_xor = int(a[i+1], 16)
			elif a[i] == b'addend':
				self.map_addend = int(a[i+1], 16)

			elif a[i] == b'rshift':
				self.map_rshift = int(a[i+1])

			elif a[i] == b'keys':
				keys = a[i+1].split(b',')
				self.keys = list(map(lambda x: x.decode('ascii'), keys))
		return self.handle

	def __str__(self):
		if not hasattr(self, 'handle') or not self.handle:
			return super(type(self), self).__str__()
		m = "handle 0x{0:x}".format(self.handle)
		if self.mode == 'hash':
			m = m + " hash keys {0}".format(self.keys)
			m = m + " divisor {0}".format(self.divisor)
			if hasattr(self, "perturb"):
				m = m + " perturb {0}sec".format(self.perturb)
		elif self.mode == 'map':
			m = m + " map keys {0}".format(self.keys)
			m += " and 0x{0:x}".format(self.map_and)	if hasattr(self, 'map_and') else ''
			m += " or 0x{0:x}".format(self.map_or)		if hasattr(self, 'map_or') else ''
			m += " xor 0x{0:x}".format(self.map_xor)	if hasattr(self, 'map_xor') else ''
			m += " addend 0x{0:x}".format(self.map_addend)	if hasattr(self, 'map_addend') else ''
			m += " rshift {0}".format(self.map_rshift)	if hasattr(self, 'map_rshift') else ''

		if hasattr(self, "baseclass"):
			m = m + " baseclass {0}".format(self.baseclass)
		if hasattr(self, "divisor"):
			m = m + " divisor {0}".format(self.divisor)
		if hasattr(self, "classid"):
			m = m + " classid {0}".format(self.classid)
		return m

	def add_filter(self, a):
		child = TcFilterFlow(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		self.children[handle] = child
		return child

	def add_line(self, l):
		self.probably_ematch(l)

	def identify(self):
		if hasattr(self, 'handle') and self.handle:
			return self.handle
		return self.pref

class TcFilterFw(TcFilter):
	type = "fw"
	def parameters(self, a):
		self.indev = self.handle = None	# if it's the root filter node
		for i in range(0, len(a)):
			# handle can contain "/mask" to mask the skb->mark with
			if a[i] == b'handle':
				h_m = list(map(lambda x: int(x, 16), a[i+1].split(b'/')))
				if len(h_m) == 2:
					(self.handle, self.mask) = h_m
				else:
					self.handle = h_m[0]
					self.mask = 0
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))
			elif a[i] == b'input' and a[i+1] == b'dev':
				self.indev = a[i+2].decode('ascii')
		return self.handle
		
	def __str__(self):
		if not hasattr(self, 'handle') or not self.handle:
			return super(type(self), self).__str__()
		m = "handle 0x{0:x}".format(self.handle)
		if hasattr(self, 'mask'):
			m += '/0x{0:x}'.format(self.mask)

		if hasattr(self, 'classid'):
			m += ' classid {0}'.format(self.classid)
		if hasattr(self, 'indev'):
			m += ' input dev {0}'.format(self.indev)
		return m

	def add_filter(self, a):
		child = TcFilterFw(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		self.children[handle] = child
		return child

	def identify(self):
		if hasattr(self, 'handle') and self.handle:
			return self.handle
		return self.pref

class TcFilterRoute(TcFilter):
	type = "route"
	def parameters(self, a):
		self.fh = None	# if it's the root filter node
		for i in range(0, len(a)):
			if a[i] == b'fh':
				self.fh = int(a[i+1], 16)
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))
			elif a[i] == b'fromif':
				self.fromif = a[i+1].decode('ascii')
			elif a[i] == b'from':
				self.fromrealm = int(a[i+1])
			elif a[i] == b'to':
				self.torealm = int(a[i+1])
		return self.fh

	def __str__(self):
		if not hasattr(self, 'fh') or not self.fh:
			return super(type(self), self).__str__()
		m = "route"
		if hasattr(self, 'fromif'):
			m += ' fromif {0}'.format(self.fromif)
		if hasattr(self, 'fromrealm'):
			m += ' from {0}'.format(self.fromrealm)
		if hasattr(self, 'torealm'):
			m += ' to {0}'.format(self.torealm)

		if hasattr(self, 'classid'):
			m += ' classid {0}'.format(self.classid)
		return m

	def add_filter(self, a):
		child = TcFilterRoute(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		self.children[handle] = child
		return child

	def identify(self):
		if hasattr(self, 'fh') and self.fh:
			return self.fh
		return self.pref

class TcFilterTcindex(TcFilter):
	type = "tcindex"
	def parameters(self, a):
		self.handle = self.policer = None
		self.fall_through = False
		for i in range(0, len(a)):
			if a[i] == b'hash':
				self.hash = int(a[i+1])
			elif a[i] == b'mask':
				self.mask = int(a[i+1], 16)
			elif a[i] == b'shift':
				self.shift = int(a[i+1])
			elif a[i] == b'fall_through':
				self.fall_through = True
			elif a[i] == b'pass_on':
				self.fall_through = False

			elif a[i] == b'handle':
				self.handle = int(a[i+1], 16)
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))
		return self.handle

	def __str__(self):
		if not hasattr(self, 'hash') and not hasattr(self, 'classid'):
			return super(type(self), self).__str__()

		# first level should print out pref/proto as well.
		if hasattr(self, 'hash'):
			m = 'filter tcindex [{0}/{1}]'.format(self.proto, self.pref)
			m += ' hash {0} mask 0x{1:x} shift {2}'.format(self.hash, self.mask, self.shift)
		# deep down, just plain information
		if hasattr(self, 'classid'):
			m = 'handle 0x{0:x} classid {1}'.format(self.handle, self.classid)
		ft = 'fall_through' if self.fall_through else 'pass_on'
		m += ' {0}'.format(ft)
		return m

	def add_filter(self, a):
		child = TcFilterTcindex(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		self.children[handle] = child
		return child

	def add_line(self, l):
		""" tcindex doesn't support actions or ematches, just the policer. """
		l = l.strip()
		if l.startswith(b'police '):
			self.policer = TcActionPolice(b'0:')
			self.policer.parameters(l[len('police '):].split())

	def identify(self):
		if hasattr(self, 'classid'):
			return self.handle
		return self.pref

class TcFilterU32(TcFilter):
	type = "u32"
	def parameters(self, a):
		self.fh = None	# if it's the root filter node
		for i in range(0, len(a)):
			if a[i] == b'fh':
				self.fh = a[i+1]
			elif a[i] == b'flowid' or a[i] == b'classid':
				self.classid = majmin_encode(*majmin_decode(a[i+1]))
			elif a[i] == b'divisor':
				self.divisor = int(a[i+1])
			elif a[i] == b'link':
				self.link = majmin_encode(*majmin_decode(a[i+1]))
				self.mask = self.shift = self.plus = None
				self.offoff = None
				self.eat = False

		if not self.fh:	# the root node for the whole filter
			return None
		numbers = list(majmin_decode(self.fh))
		if len(numbers) == 3:
			self.htid, self.bkt, self.key = numbers
		else:
			self.htid, self.bkt = numbers
			self.key = 0
		return majmin_encode(self.htid, self.bkt, self.key)

	def __str__(self):
		if not hasattr(self, 'fh') or not self.fh:
			return super(type(self), self).__str__()
		m = "fh {0} ".format(self.fh.decode('ascii'))
		if hasattr(self, "divisor"):
			m = m + "divisor {0}".format(self.divisor)
		elif hasattr(self, "mask") and self.mask:
			m = m + "link {0} mask {1:08x} >> {2} at {3}".format(self.link, self.mask, self.shift, self.offoff)
			if self.plus:
				m = m + " plus {0}".format(self.plus)
			if self.eat:
				m = m + " eat"
		elif hasattr(self, "hashmask") and self.hashmask:
			m = m + "link {0} hash mask {1:08x} at {2}".format(self.link, self.hashmask, self.hashoff)
		elif hasattr(self, "classid"):
			m = m + "classid {0}".format(self.classid)
		return m

	def add_filter(self, a):
		child = TcFilterU32(self.pref, self.proto, self.parent)
		handle = child.parameters(a)
		if child.htid not in self.children:
			if child.key:
				print("U32 order wrong: {0}".format(child))
			self.children[child.htid] = child
		else:
			k = (child.bkt << 12) | child.key
			self.children[child.htid].children[k] = child
		return child

	def add_line(self, line):
		f = line.split()
		if f[0] == b'match' and f[2] == b'at':
			[key, mask] = list(map(lambda x: int(x, 16), f[1].split(b"/")))
			if f[3].startswith(b"nexthdr+"):
				off = int(f[3][8:]) | (1 << 16)
			else:
				off = int(f[3])
			self.matches = self.matches + [(key, mask, off)]
		elif f[0] == b'offset' and f[2] == b'at':
			mask, shift = f[1].split(b'>>')
			self.mask = int(mask, 16)
			self.shift = int(shift)
			self.offoff = int(f[3])
			f = f[4:]
			while len(f) > 0:
				if f[0] == b'eat':
					self.eat = True
					f = f[1:]
				elif f[0] == b'plus':
					self.plus = int(f[1])
					f = f[2:]
		elif f[0] == b'hash' and f[1] == b'mask' and f[3] == b'at':
			self.hashmask = int(f[2], 16)
			self.hashoff = int(f[4])
		else:
			raise Exception("U32 filter unknown line: >>{0}<<".format(line))

	def identify(self):
		if hasattr(self, 'fh') and self.fh:
			if self.key:
				return (self.bkt << 12) | self.key
			return self.htid
		return self.pref


def TcParseFilters(s):
	"""Parses the output of "tc filter show"."""
	fls = dict()
	last = lastfilter = None
	for line in s.split(b'\n'):
		if not line:
			continue
		elif line.startswith(b'filter '):
			f = line.split()
			if len(f) < 6:
				raise Exception("short line: {0}".format(line))

			# BAD_BAD_TC dsmark classes are showing their parent's shit
			if f[1] == b'parent':
				parent = majmin_encode(majmin_decode(f[2]))
				off = 2
			else:
				parent = None
				off = 0

			if f[off+1] != b'protocol' or f[off+3] != b'pref':
				raise Exception("bad line: {0}".format(line))

			(t, pref, proto) = (f[off +5], int(f[off +4]), f[off +2].decode('ascii'))

			if pref in fls:
				lastfilter = last = fls[pref].add_filter(f[off+6:])
				continue

			# Actually create the filter now, based on its type.
			# This code is a bit longer than necesarry, but it'll
			# tell us when we need to add new types of filters here.

			if t == b'basic':
				last = TcFilterBasic(pref, proto, parent)
			elif t == b'bpf':
				last = TcFilterBPF(pref, proto, parent)
			elif t == b'flow':
				last = TcFilterFlow(pref, proto, parent)
			elif t == b'fw':
				last = TcFilterFw(pref, proto, parent)
			elif t == b'route':
				last = TcFilterRoute(pref, proto, parent)
			elif t == b'tcindex':
				last = TcFilterTcindex(pref, proto, parent)
			elif t == b'u32':
				last = TcFilterU32(pref, proto, parent)
			else:
				raise Exception("bad filter type {0}".format(t))

			if last.parameters(f[off+6:]):
				raise Exception("{0} supposed2B root".format(line))
			fls[last.pref] = lastfilter = last

		elif line.strip().startswith(b'action'):
			f = line.split()
			if f[1] != b'order':
				m = "action not in order, {0}".format(line)
				raise Exception(m)
			last = TcNewAction(f[3], f[2])
			if last.parameters(f[4:]):
				raise Exception("{0} weird action".format(line))
			lastfilter.add_action(last)

		elif line.strip().startswith(b'match'):
			lastfilter.add_line(line)

		else:
			last.add_line(line)
	return fls

