#!/usr/bin/env python3.3

from select import select
from time import time, sleep
from threading import Lock

from paramiko import AuthenticationException, BadHostKeyException, PasswordRequiredException, SSHException

from host import SSH_Host, ssh_key_digest

class QViz:
	def __init__(self, fe_module='tkinter'):
		from importlib import import_module
		fe = import_module('frontends.' + fe_module)
		self.frontend = fe.current_frontend(self)

		self.hosts = dict()
		self.hosts_to_add = dict()
		self.hosts_to_delete = []
		self.hosts_lock = Lock()
		self.hosts_changes_lock = Lock()

		from os.path import expanduser, join
		self.known_hosts = join(expanduser('~'), ".ssh", "known_hosts")

		from os import mkdir, stat
		try:	stat(self.known_hosts)
		except:
			try: mkdir(join(expanduser('~'), ".ssh"))
			except: pass
			try:
				f = open(self.known_hosts, "w")
				f.close()
			except IOError as ioe:
				m = "Can't open {0}, {1}".format(self.known_hosts, ioe)
				self.frontend.show_error(self.known_hosts, m)

		# unconfirmed SSH keys can go here and wait for the user
		self.quarantine = dict()

	def quit(self):
		self.frontend.wanna_run = False

	def process_events(self, r=[]):
		"""Called from the event loop, so it'd better be fast."""

		# XXX BAD_BAD_TKINTER does its events whenever it feels like it
		# XXX http://bugs.python.org/issue11387 it is.

		# Here's the locking protocol:
		# Everyone iterating over self.hosts acquires self.hosts_lock.
		# Should anyone add anything, give it to self.hosts_to_add.
		# After we finish iterating, release self.hosts_lock, grab
		# self.hosts_changes_lock, switch hosts_to_add for an empty
		# one, unlock it, grab self.hosts_lock and add them all here.
		#
		# The key invariant is obviously that both locks mustn't be
		# taken simultaneously at any given moment except manipulating
		# self.hosts_to_{add, delete}, which are only used when the
		# main thing is locked.

		rlist = []
		wlist = []
		ext_read = dict()
		for (addr, fn) in r:
			# type unchecked, dirty hack XXX
			if addr not in self.frontend.connections:
				sock = addr
			else:
				sock = self.frontend.connections[addr]
			rlist += [sock]
			ext_read[sock] = (fn, addr)

		rhosts = dict()
		whosts = dict()
		self.hosts_lock.acquire()
		for hk in self.hosts:
			h = self.hosts[hk]
			if h.connected and h.session:
				if h.doing_upload:
					# XXX again, more testing needed.
					# this should be safe, but slow
					if h.session.send_ready():
						self.frontend.log(hk + " send ready")
						wlist = wlist + [h.session]
						whosts[h.session] = h
					#else:
					#	self.frontend.log(hk + " send NOT ready")
				else:
					#if h.session.recv_ready():
					#	self.frontend.log(hk + " recv ready")
					rlist = rlist + [h.session]
					rhosts[h.session] = h
			elif not h.connected:
				self.fix_connection(h)
				#self.frontend.log_host_status(h)
			elif not h.curtask:
				while not h.session:
					if not h.periodic_events(self.frontend):
						break
					if h.tasks_in_queue() == 0:
						break
					if not h.cat_alive():
						break
					h.curtask = h.pop_task()
					if not h.curtask:
						break
					(cmd, _, sudo) = h.curtask
					#m = "scheduled {0} with {1}".format(h, cmd)
					#self.frontend.log(m)
					if sudo:
						h.sudo(cmd)
					else:
						h.run(cmd)
				if h.session:
					#if h.session.recv_ready():
					#	self.frontend.log(hk + " recv ready")
					rlist = rlist + [h.session]
					rhosts[h.session] = h
		self.hosts_lock.release()

		self.hosts_changes_lock.acquire()
		towipe = self.hosts_to_delete
		toadd = self.hosts_to_add
		self.hosts_to_add = dict()
		self.hosts_to_delete = []
		self.hosts_changes_lock.release()

		self.hosts_lock.acquire()
		for h in towipe:
			hh = self.hosts[h]
			self.frontend.ltv_delete_node(hh.widget)
			hh.widget = None
			del self.hosts[h]

		for h in toadd:
			self.hosts[h] = toadd[h]
			# It is safe to show the host now.
			self.frontend.add_host(toadd[h])
		self.hosts_lock.release()

		# windoze can't do empty select()
		if len(rlist) == 0 and len(wlist) == 0:
			sleep(.1)
		else:
			#self.frontend.log("rlist: "+str(rlist))
			#self.frontend.log("wlist: "+str(wlist))
			(rl, wl, xl) = select(rlist, wlist, [], .2)
		for s in rhosts:
			h = rhosts[s]
			if not h.connected and h.waiting_too_long():
				msg = "{0}: retrying".format(h.hostname)
				self.frontend.log(msg)
				h.close()
				self.fix_connection(h)
			elif s in rl:
				if h.move_forward():
					# only refresh, if we're selected
					self.frontend.ltv_refresh(h.widget)

		for s in whosts:
			h = whosts[s]
			if s in wl:
				self.frontend.log("{0} can write".format(h))
				h.move_forward(can_write=True)
			else:
				self.frontend.log("{0} can't write".format(h))
				h.check_if_exited()

		# Whatever the frontend can throw at us.
		for sock in ext_read:
			if sock in rl:
				(fn, addr) = ext_read[sock]
				fn(addr, sock)
		del rlist
		del rhosts
		del wlist
		del whosts

	def fix_connection(self, h):
		newfd = None
		try:
			newfd = h.connect()
		# XXX TODO FIXME etc etc.
		# This is a brilliant example of how subtle can bad code be,
		# in terms of sticking to a design.
		# the AuthExc is done properly, leaving the frontend actions
		# to us (the caller), whereas BadHostKeyExc does some magic
		# hooked by the frontend (which has no choice of notification
		# because it can't see other parts of the program) and we get
		# to do only its cleanup (print message saying it's too late
		# and disconnect).
		# XXX set this thing right!
		except BadHostKeyException as bhke:
			cipher = bhke.key.get_name()
			d = ssh_key_digest(h.expected_key(cipher))
			h.set_status("{0} Expected {1} {2}".format(bhke, cipher, d))
			self.disconnect_host(h.name)
		except AuthenticationException as ae:
			m = "{0}\n{1}".format(h.name, ae)
			p = self.frontend.ask_for_password(m)
			if not p:
				self.disconnect_host(h.name)
			else:
				h.password = p
			h.set_status("Authentication failed: "+str(ae))
		except SSHException as se:
			h.set_status("SSH Exception: "+str(se))
			self.disconnect_host(h.name)

		return newfd

	# Gets called by SOMEONE and needs to call the frontend back.
	# We can be on a large display controlled by some CLI tool remotely.

	def add_host(self, config):
		if 'hostname' not in config or not config['hostname']:
			self.frontend.log("can't add host without hostname!")
			return
		hostname = config['hostname']
		try:	port = int(config['port'])
		except:	port = 22
		name = config['name'] if 'name' in config else hostname

		h = SSH_Host(name)
		h.set_logging_function(self.frontend.log)
		h.resolve_hostname(hostname, port)
		h.set_username_password(config['user'], config['password'])
		h.set_password_dialog(self.frontend.ask_for_password)
		h.load_host_keys(self.known_hosts)
		h.set_pubkey_policy(self.frontend_key_policy)
		self.frontend.log_host_status(h)

		if 'identityfile' in config:
			p = None
			for f in config['identityfile']:
				try:	h.set_identityfile(f, password=p)
				except PasswordRequiredException as pre:
					m = "{0}\n{1}".format(f, pre)
					p = self.frontend.ask_for_password(m)
				try:	h.set_identityfile(f, password=p)
				except:	pass

		# Add the new host to our global list of hosts.
		if not self.hosts_lock.acquire(False):
			self.hosts_changes_lock.acquire()
			self.hosts_to_add[name] = h
			self.hosts_changes_lock.release()
		else:
			self.hosts[name] = h

			# Frontend requires it to be on the main list.
			# What if'd someone clicked on it while still
			# in our hosts_to_add quarantine?
			self.frontend.add_host(h)

			self.hosts_lock.release()

	def disconnect_host(self, host):
		if host not in self.hosts:
			return
		h = self.hosts[host]
		if h.connected:
			h.disconnect()

		# Delete the host from our global list of hosts.
		if not self.hosts_lock.acquire(False):
			self.hosts_changes_lock.acquire()
			self.hosts_to_delete += [host]
			self.hosts_changes_lock.release()
		else:
			del self.hosts[host]

			self.frontend.ltv_delete_node(h.widget)
			self.frontend.disconnect_host(h.name, ui_origin=False)
			h.widget = None
			self.hosts_lock.release()

	def refresh_host(self, host, use_sudo=False):
		if host not in self.hosts:
			raise Exception("error, host not registered!")

		h = self.hosts[host]
		if not h.connected:
			if self.fix_connection(h):
				self.frontend.log_host_status(h)

		if h.locked:
			raise Exception("operation {0} in progress".format(h.lastcmd))
		if not h.connected:
			self.frontend.log("retrying {0}".format(h))
			h.resolve_hostname(h.hostname, h.port)
			self.fix_connection(h)

		h.locked = True
		for sw in h.software:
			for (st, fn) in sw.subtrees():
				fn(self.frontend, use_sudo)
		h.locked = False

	def run(self):
		self.frontend.run()

if __name__ == '__main__':
	q = QViz()
	#q.frontend.daemonize()
	q.frontend.run()
	q.quit()

