
# CentOS has it in /sbin, normal systems do in /bin.  So there.
IP_PATH = "ip"
TC_PATH = "/sbin/tc"

from collections import deque

from parsers.tc import TcFilter, TcParseLinks, TcParseQdiCls, TcParseFilters
from software import QVizSoftware
from software.Ubiquiti import Ubiquiti

class Linux(QVizSoftware):
	name = "Linux"

	def __init__(self):
		self.links = dict()
		self.tc_refcnt = 0
		self.allifqdiscs = dict()
		self.allifclasses = dict()

	def node_types(self):
		return ['interface', 'qdisc', 'class', 'filter']

	def subtrees(self):
		return [('tc_tree', self.refresh_tc_tree)]

	def refresh_tc_tree(self, frontend, use_sudo=False):
		self.use_sudo = use_sudo
		self.async_fetch_links(frontend)
		self.async_fetch_tctree(frontend, startit=True)

	def whatis(self, path):
		for [k, h] in path:
			if k == 'interface':
				dev = h
				handle = self.links[h]
				t = ["interface"]
			elif k == 'qdisc':
				handle = self.allifqdiscs[dev][h]
				t = ["qdisc", handle['type']]
			elif k == 'class':
				handle = self.allifclasses[dev][h]
				t = ["class", handle['type']]
			elif k == 'filter':
				if isinstance(handle, TcFilter):
					handle = handle.children[h]
				else:
					handle = handle['filters'][h]
				t = ["filter", handle.type]
			elif k == 'host':
				t = ["host"]
				handle = self.host
			else:
				m = "Linux: weird k {0} h {1}".format(k, h)
				self.host.log(m)
				break
		return (t, handle)

	def fetch_links(self):
		if self.use_sudo:
			(s, e) = self.host.sudo(IP_PATH + ' link show')
		else:
			(s, e) = self.host.run(IP_PATH + ' link show')
		if e != "":
			self.host.log(e)
		nlinks = TcParseLinks(s)
		# just copy the old interfaces' widgets over, unifying them
		for l in nlinks:
			if l in self.links:
				nlinks[l]['widget'] = self.links[l]['widget']
				nlinks[l]['egress'] = self.links[l]['egress']
				nlinks[l]['ingress'] = self.links[l]['ingress']
		return nlinks

	def tcdesc(self, node):
		if 'qdcl' in node and node['qdcl'] == 'qdisc':
			qd = "qdisc {0} {1} ({2})".format(node['handle'], \
			    node['type'], node['rest'].decode('ascii'))
			return qd
		elif 'qdcl' in node and node['qdcl'] == 'class':
			cl = "class {0} {1} ({2})".format(node['handle'], \
			    node['type'], node['rest'].decode('ascii'))
			return cl
		else:
			raise Exception("can't make a description to {0}".format(node))

	def add_subtree(self, parwidget, node, fe, myhandle=None):
		desc = self.tcdesc(node)
		# BAD_BAD_TC	The key to get to this node might be different
		# from its handle.  We'll try our best to pass them here.
		if not myhandle:
			myhandle = node['handle']

		(t, id) = (node['qdcl'], myhandle)
		w = fe.ltv_add_node(parwidget, desc, t, id)
		node['widget'] = w
		for qkey in sorted(node['qdiscs']):
			q = node['qdiscs'][qkey]
			self.add_subtree(w, q, fe, qkey)
		for ckey in sorted(node['classes']):
			c = node['classes'][ckey]
			self.add_subtree(w, c, fe, ckey)
		for pref in sorted(node['filters']):
			node['filters'][pref].update_frontend(fe, w)

	def merge_filters(self, ol, nw, fe):
		of = ol.children if isinstance(ol, TcFilter) else ol['filters']
		nf = nw.children if isinstance(nw, TcFilter) else nw['filters']
		for k in of:
			if k in nf:
				# copy over and sync the good ones
				nf[k].widget = of[k].widget
				self.merge_filters(of[k], nf[k], fe)
			else:
				# remove unused filters
				of[k].delete_frontend(fe)

	def zipcls(self, ol, nw):
		"""ol/nw -> (to_delete, to_add, zip(to_update))"""
		q2del = [q for q in ol['qdiscs'] if q not in nw['qdiscs']]
		c2del = [c for c in ol['classes'] if c not in nw['classes']]
		todel = [ol['qdiscs'][qk] for qk in q2del]
		todel = todel + [ol['classes'][ck] for ck in c2del]

		q2add = [q for q in nw['qdiscs'] if q not in ol['qdiscs']]
		c2add = [c for c in nw['classes'] if c not in ol['classes']]
		toadd = [nw['qdiscs'][qk] for qk in q2add]
		toadd = toadd + [nw['classes'][ck] for ck in c2add]

		q2upd = sorted([q for q in nw['qdiscs'] if q in ol['qdiscs']])
		c2upd = sorted([c for c in nw['classes'] if c in ol['classes']])

		oldqd = [ol['qdiscs'][qk] for qk in q2upd]
		newqd = [nw['qdiscs'][qk] for qk in q2upd]
		oldcl = [ol['classes'][ck] for ck in c2upd]
		newcl = [nw['classes'][ck] for ck in c2upd]

		toupd = zip(oldqd + oldcl, newqd + newcl)
		return (todel, toadd, toupd)

	def tctree_merge(self, fe, newroots):
		"""Show all changes to all interfaces on this host in frontend."""
		for ifname in newroots:
			if ifname not in self.links:	# not yet loaded up
				continue
			link = self.links[ifname]
			(root, ingress) = newroots[ifname]
			self.tctree_merge_interface(link, root, 'egress', fe)
			self.tctree_merge_interface(link, ingress, 'ingress', fe)
				
		del newroots

	def tctree_merge_interface(self, link, nstart, direction, fe):
		"""BFS over both trees and fix up differences in frontend."""
		if len(link[direction]) == 0:
			link[direction] = nstart
			if len(nstart):
				self.add_subtree(link['widget'], nstart, fe)
			return

		# Imagine the whole ingress tree disappears.
		if len(nstart) == 0:
			fe.ltv_delete_node(link[direction]['widget'])
			link[direction] = nstart
			return

		q = deque()
		q.append((link[direction], nstart))
		while len(q) > 0:
			(ol, nw) = q.popleft()

			(oroot, nroot) = ('parent' not in ol,'parent' not in nw)
			opar = ol['parent']['handle'] if not oroot else None
			npar = nw['parent']['handle'] if not nroot else None

			(oqc, nqc) = (ol['qdcl'], nw['qdcl'])
			(oty, nty) = (ol['type'], nw['type'])
			(oha, nha) = (ol['handle'], nw['handle'])
			(orest, nrest) = (ol['rest'], nw['rest'])

			if oqc == nqc and opar == npar and \
			    oty == nty and oha == nha:
				w = nw['widget'] = ol['widget']
				desc = self.tcdesc(nw)
				fe.ltv_update_node(w, desc, nqc, nha)
				self.merge_filters(ol, nw, fe)
				for fk in nw['filters']:
					f = nw['filters'][fk]
					f.update_frontend(fe, w)

				# returns which ones to keep or delete
				# and zipped pairs to check further
				(de, ad, up) = self.zipcls(ol, nw)
				q.extend(up)
				for d in de:
					fe.ltv_delete_node(d['widget'])
				for a in ad:
					self.add_subtree(w, a, fe)
			else:
				if 'widget' in ol:
					fe.ltv_delete_node(ol['widget'])
				parwidget = link['widget'] if nroot else \
				    nw['parent']['widget']
				self.add_subtree(parwidget, nw, fe)
		link[direction] = nstart

	def tctree_connect(self):
		nd = dict()
		# Note that x['parent'] was a string until now; will be dict().
		# This enables us to do tree traversal backwards when searching.
		# (It's a shame that type info must be in the comments, innit?)

		for ifname in self.allifqdiscs:
			qifc = self.allifqdiscs[ifname]
			if ifname in self.allifclasses:
				cifc = self.allifclasses[ifname]
			else:
				cifc = dict()

			roots = [q for q in qifc if 'parent' not in qifc[q]]
			if len(roots) > 2:
				raise Exception("{0} roots: {1}".format(ifname, roots))
			elif len(roots) == 2:
				a = self.allifqdiscs[ifname][roots[0]]
				b = self.allifqdiscs[ifname][roots[1]]
				if a['type'] == 'ingress':
					root = roots[1]
					ingress = roots[0]
				elif b['type'] == 'ingress':
					root = roots[0]
					ingress = roots[1]
				else:
					raise Exception("two roots weird")
			elif len(roots) == 1:
				root = roots[0]
				ingress = None
			else:
				return None

			for qkey in qifc:
				q = qifc[qkey]
				if 'parent' not in q:
					continue

				# I'm a child qdisc, link me up to my parent!
				h = q['handle']
				parhdl = q['parent']
				if parhdl in qifc:	# my dad's a qdisc
					qifc[parhdl]['qdiscs'][qkey] = q
					q['parent'] = qifc[parhdl]
				elif parhdl in cifc:	# my mum's a class
					cifc[parhdl]['qdiscs'][qkey] = q
					q['parent'] = cifc[parhdl]
			for ckey in cifc:
				c = cifc[ckey]
				if 'parent' not in c:
					continue

				# I'm a child class, link me up to my parent!
				h = c['handle']
				parhdl = c['parent']
				if parhdl in qifc:	# my dad's a qdisc
					qifc[parhdl]['classes'][ckey] = c
					c['parent'] = qifc[parhdl]
				elif parhdl in cifc:	# my mum's a class
					cifc[parhdl]['classes'][ckey] = c
					c['parent'] = cifc[parhdl]
			nd[ifname] = (qifc[root], qifc[ingress] if ingress else {})
		return nd

	def async_fetch_links(self, frontend):
		self.cur_frontend = frontend
		return self.host.push_task(IP_PATH + ' link show', self.finish_fetch_links)

	def finish_fetch_links(self, o, e=""):
		fe = self.cur_frontend

		if e != "":
			self.host.log(e)
		nlinks = TcParseLinks(o)

		# just copy the old interfaces' widgets over, unifying them
		for l in nlinks:
			if l in self.links:
				nlinks[l]['widget'] = self.links[l]['widget']
				nlinks[l]['egress'] = self.links[l]['egress']
				nlinks[l]['ingress'] = self.links[l]['ingress']

		# wipe the old ones
		for i in self.links:
			if i in nlinks:
				continue
			w = self.links[i]['widget']
			fe.ltv_delete_node(w)

		# make them interfaces go away
		self.links = nlinks

		# update or add new ones
		for i in sorted(nlinks):
			l = nlinks[i]
			desc = "{0}: {1}".format(i, l['flags'])
			t = "interface"
			if 'widget' in l:
				w = l['widget']
				fe.ltv_update_node(w, desc, t, i)
			else:
				w = fe.ltv_add_node(self.host.widget, desc, t, i)
				l['widget'] = w

	def async_fetch_tctree(self, frontend, startit=False):
		cmd = TC_PATH + ' qdisc show'
		self.host.log("running {0}".format(cmd))
		self.frontend = frontend
		return self.host.push_task(cmd, self.finish_fetch_qdiscs, \
		    use_sudo=self.use_sudo)

	def finish_fetch_qdiscs(self, o, e=None):
		# parse the output
		self.allifqdiscs = TcParseQdiCls(o)
		self.tc_refcnt = 0
		for iface in self.allifqdiscs:
			cmd = TC_PATH + ' class show dev {0}'.format(iface)
			self.host.push_task(cmd, self.finish_fetch_class, \
			    use_sudo=self.use_sudo)
			self.tc_refcnt = self.tc_refcnt + 1

			for qd in self.allifqdiscs[iface]:
				# BAD_BAD_TC	QFQ won't show any classes
				# unless we do a pretty fucking please
				# (with a strawberry and cheese on top).
				if self.allifqdiscs[iface][qd]['type'] == 'qfq':
					cmd = TC_PATH + ' class show dev {0} parent {1}'
					realname = self.allifqdiscs[iface][qd]['handle']
					cmd = cmd.format(iface, realname)
					self.host.push_task(cmd, self.finish_fetch_class, \
					    use_sudo=self.use_sudo)
					self.tc_refcnt = self.tc_refcnt + 1

				# BAD_BAD_TC	To honor tc's stupidity,
				# we can't assume the array key to be the
				# actual handle.  pfifo_fast 0: under mq.
				cmd = TC_PATH + ' filter show dev {0} parent {1}'
				realname = self.allifqdiscs[iface][qd]['handle']
				cmd = cmd.format(iface, realname)
				self.host.push_task(cmd, self.finish_fetch_filter, \
				    use_sudo=self.use_sudo)
				self.tc_refcnt = self.tc_refcnt + 1

				# Now try to make sense of it.
				self.parse_qdisc(iface, qd)
		self.check_tc_refcnt_merge()

	def check_tc_refcnt_merge(self):
		if self.tc_refcnt == 0:
			self.host.log("merging tc tree...")
			nd = self.tctree_connect()
			if nd:
				self.tctree_merge(self.frontend, nd)
				self.host.log("...done!")
			else:
				self.host.log("...failed!")
			self.cur_frontend = None
			self.locked = False
			return

	def finish_fetch_class(self, o, e=None):
		self.tc_refcnt = self.tc_refcnt - 1

		oldcmd = self.host.lastcmd.split()
		parent = None
		for i in range(0, len(oldcmd)):
			if oldcmd[i] == 'dev':
				dev = oldcmd[i+1]
			elif oldcmd[i] == 'parent':
				parent = oldcmd[i+1]

		# BAD_BAD_TC	In case buggy disciplines like QFQ
		# decide we have to query their classes separately.
		newstuff = TcParseQdiCls(o, ifname=dev)
		if not parent:
			self.allifclasses[dev] = newstuff
		else:
			for k in newstuff:
				assert(k not in self.allifclasses[dev])
				self.allifclasses[dev][k] = newstuff[k]

		for ck in newstuff:
			c = self.allifclasses[dev][ck]
			cmd = TC_PATH + ' filter show dev {0} parent {1}'
			cmd = cmd.format(dev, c['handle'])
			self.host.push_task(cmd, self.finish_fetch_filter, \
			    use_sudo=self.use_sudo)
			self.tc_refcnt = self.tc_refcnt + 1

			# Now try to make sense of it.
			self.parse_class(dev, ck)
		self.check_tc_refcnt_merge()

	def finish_fetch_filter(self, o, e=None):
		self.tc_refcnt = self.tc_refcnt - 1

		oldcmd = self.host.lastcmd.split()
		for i in range(0, len(oldcmd)):
			if oldcmd[i] == 'dev':
				dev = oldcmd[i+1]
			elif oldcmd[i] == 'parent':
				parent = oldcmd[i+1]

		try:
			filters = TcParseFilters(o)
		except Exception as ex:
			self.host.log("parsing filters "+str(ex))
			self.check_tc_refcnt_merge()
			return
		haveqd = dev in self.allifqdiscs
		havecl = dev in self.allifclasses
		if haveqd and parent in self.allifqdiscs[dev]:
			parent_node = self.allifqdiscs[dev][parent]
		elif havecl and parent in self.allifclasses[dev]:
			parent_node = self.allifclasses[dev][parent]
		else:
			self.host.log("No parent node, aborting")
			self.check_tc_refcnt_merge()
			return

		# BAD_BAD_TC dsmark classes acting up again, showing bad parent:
		pref2delete = []
		for pref in filters:
			npar = filters[pref].parent
			# if tc(8) didn't bother showing, just stick it to ours
			if not npar:
				npar = parent
			if npar != parent:
				# reparent these filters to their proper parents
				if haveqd and npar in self.allifqdiscs[dev]:
					q = self.allifqdiscs[dev][npar]
					q['filters'][pref] = filters[pref]
					pref2delete = pref2delete + [pref]
				elif havecl and npar in self.allifclasses[dev]:
					c = self.allifclasses[dev][npar]
					c['filters'][pref] = filters[pref]
					pref2delete = pref2delete + [pref]
			else:
				# stick the filter to its parent from lastcmd
				parent_node['filters'][pref] = filters[pref]
		for pref in pref2delete:
			del filters[pref]
		self.check_tc_refcnt_merge()

	def parse_qdisc(self, dev, qd):
		from parsers.tc import TcParseHFSC

		q = self.allifqdiscs[dev][qd]
		if q['type'] == 'hfsc':
			curves = TcParseHFSC(q['rest'])
			for k in curves:
				q[k] = curves[k]

	def parse_class(self, dev, cl):
		from parsers.tc import TcParseHFSC, TcParseHTB

		c = self.allifclasses[dev][cl]
		if c['type'] == 'hfsc':
			curves = TcParseHFSC(c['rest'])
			for k in curves:
				c[k] = curves[k]
		elif c['type'] == 'htb':
			htb = TcParseHTB(c['rest'])
			for k in htb:
				c[k] = htb[k]

	def linux_based_on_etc(self, o, e=None):
		files = o.split()
		h = self.host
		# we might look for centos-release, gentoo-release, ...
		if b'airview.conf' in files:
			u = Ubiquiti()
			if not h.register_software(u):
				h.log("{0}: can't register {1}".format(h.name, u))
				return
			u.host = h
			h.log("{0}: registered {1}".format(h.name, u))

