from collections import deque

from parsers.bsd import BSDParseLinks, BSDParseALTQ, BSDParseHFSC, BSDParseNewQueue
from software import QVizSoftware

class OpenBSD(QVizSoftware):
	name = "OpenBSD"

	cmd_anchors = '/sbin/pfctl -sA '
	cmd_rules = '/sbin/pfctl -sr '

	def __init__(self):
		self.links = dict()
		self.altqs = dict()
		self.rulesets = dict()

	def node_types(self):
		return ['interface']

	def subtrees(self):
		return [('altq', self.refresh_altq), ('pf', self.refresh_pf)]

	def refresh_altq(self, frontend, use_sudo=False):
		self.use_sudo = use_sudo
		self.async_fetch_links(frontend)
		self.async_fetch_queues(frontend)

	def refresh_pf(self, frontend, use_sudo=False):
		self.use_sudo = use_sudo
		self.async_fetch_rulesets(frontend)

	def whatis(self, path):
		for [k, h] in path:
			if k == 'interface':
				dev = h
				handle = self.links[h]
				t = ["interface"]
			elif k == 'host':
				handle = self.host
				t = ["host"]
			elif k == 'queue':
				handle = self.altqs[dev][h]
				t = ["queue"]
			elif k == 'rules':
				handle = self.rulesets[h]['rules']
				t = ["rules"]
			else:
				break
		return (t, handle)

	def iface_description(self, ifname, ifdict):
		"""Label for an interface @ifname with parameters in dict @ifdict."""
		desc = "{0}: ".format(ifname)
		if 'description' in ifdict:
			desc += ifdict['description'].decode('utf-8')
		if 'flags' in ifdict:
			desc += ': ' + ifdict['flags'].decode('utf-8')
		return desc

	def async_fetch_links(self, frontend):
		self.cur_frontend = frontend
		return self.host.push_task('/sbin/ifconfig', self.finish_fetch_links, \
		    self.use_sudo)

	def finish_fetch_links(self, o, e=""):
		fe = self.cur_frontend

		if e != "":
			fe.log(self.host + ': ' + str(e))
		nlinks = BSDParseLinks(o)

		# just copy the old interfaces' widgets over, unifying them
		for l in nlinks:
			if l in self.links:
				nlinks[l]['widget'] = self.links[l]['widget']

		# wipe the old ones
		for i in self.links:
			if i in nlinks:
				continue
			w = self.links[i]['widget']
			fe.ltv_delete_node(w)

		# make them interfaces go away
		self.links = nlinks

		# update or add new ones
		for i in sorted(nlinks):
			l = nlinks[i]
			desc = self.iface_description(i, l)
			t = "interface"
			if 'widget' in l:
				w = l['widget']
				fe.ltv_update_node(w, desc, t, i)
			else:
				w = self.host.widget
				l['widget'] = fe.ltv_add_node(w, desc, t, i)

	def async_fetch_queues(self, frontend):
		cmd = '/sbin/pfctl -sq'
		fn = self.finish_fetch_altqs if self.queueing == 'altq' else self.finish_fetch_queues
		return self.host.push_task(cmd, fn, use_sudo=True)

	def finish_fetch_altqs(self, o, e=""):
		fe = self.cur_frontend
		queues = BSDParseALTQ(o)
		for dev in queues:
			if dev not in self.links:
				self.host.log("wrong dev {0}".format(dev))
				continue

			self.merge_queues_interface(dev, queues[dev])

		# what if ALL of the queues were to disappear?
		for dev in self.altqs:
			if dev not in queues:
				i = self.links[dev]
				d = self.iface_description(dev, i)
				fe.ltv_delete_node(i['widget'])
				w = self.host.widget
				i['widget'] = fe.ltv_add_node(w, d, "interface", i)
		self.altqs = queues

	def finish_fetch_queues(self, o, e=""):
		fe = self.cur_frontend
		queues = BSDParseNewQueue(o)
		for dev in queues:
			if dev not in self.links:
				self.host.log("wrong dev {0}".format(dev))
				continue

			self.merge_queues_interface(dev, queues[dev])

		# what if ALL of the queues were to disappear?
		for dev in self.altqs:
			if dev not in queues:
				i = self.links[dev]
				d = self.iface_description(dev, i)
				fe.ltv_delete_node(i['widget'])
				w = self.host.widget
				i['widget'] = fe.ltv_add_node(w, d, "interface", i)
		self.altqs = queues

	def merge_queues_interface(self, dev, queues):
		fe = self.cur_frontend
		if self.queueing == 'altq':
			rootqname = "root_"+dev
			rootq = queues[rootqname]
		else:
			rootqname = [x for x in queues if 'parent' not in queues[x]][0]
			rootq = queues[rootqname]
		rootq['parent'] = None

		# Find old root's kids that are gone now and wipe them.
		# We do this separately, because we don't store the root node
		# literally and just stick the parameters to the device.
		# This is of course only if altq was on before...
		if dev in self.altqs:
			for ok in self.altqs[dev]:
				okid = self.altqs[dev][ok]
				w = okid['widget']
				if okid['parent'] == rootqname and ok not in queues:
					fe.ltv_delete_node(w)

		qq = deque()
		for k in rootq['kids']:
			queues[k]['parent'] = rootqname
			qq.append(k)
		rootq['widget'] = self.links[dev]['widget']
		if self.queueing == 'altq':
			self.links[dev]['rootq_bw'] = rootq['bandwidth']
		else:
			self.links[dev]['rootq_bw'] = rootq['linkshare'][2]

		while len(qq):
			qname = qq.popleft()
			q = queues[qname]

			pw = queues[q['parent']]['widget']
			if dev in self.altqs and qname in self.altqs[dev]:
				# find kids that are gone now and wipe them
				for ok in self.altqs[dev][qname]['kids']:
					w = self.altqs[dev][ok]['widget']
					if ok not in queues:
						fe.ltv_delete_node(w)

				# Updating label isn't necessary as it's the key.
				q['widget'] = self.altqs[dev][qname]['widget']
			else:
				w = fe.ltv_add_node(pw, qname, "queue", qname)
				q['widget'] = w

			if self.queueing != 'altq':
				for k in q['kids']:
					qq.append(k)
				continue

			if q['disc'] == 'hfsc':
				curves = BSDParseHFSC(q['params'])
				for k in curves:
					q[k] = curves[k]
				if 'kids' in q:
					for k in q['kids']:
						kid = queues[k]
						kid['parent'] = qname
						qq.append(k)
			else:
				m = "unknown " + q['disc'] + " in " + qname
				self.host.log(m)


	def async_fetch_rulesets(self, frontend):
		self.cur_frontend = frontend
		self.pf_refcnt = 2
		self.rulesets_updated = [""]

		self.host.push_task(self.cmd_anchors, self.finish_fetch_anchors, True)
		self.host.push_task(self.cmd_rules, self.finish_fetch_rules, True)

	def finish_fetch_anchors(self, o, e=""):
		fe = self.cur_frontend

		self.pf_refcnt -= 1
		if e != "":
			fe.log(self.host + ": " + str(e))
			return

		# Add all the subanchors to the queue, to be printed.
		for an in o.split(b'\n'):
			astr = an.strip().decode('ascii')
			if not astr:
				continue
			cmd = self.cmd_anchors + " -a " + astr
			self.host.push_task(cmd, self.finish_fetch_anchors, True)
			cmd = self.cmd_rules + " -a " + astr
			self.host.push_task(cmd, self.finish_fetch_rules, True)

			self.pf_refcnt += 2

	def finish_fetch_rules(self, o, e=""):
		fe = self.cur_frontend

		if e != "":
			fe.log(self.host + ": " + str(e))

		# Determine anchor's name and parent widget.
		an = self.host.lastcmd[len(self.cmd_rules) + 4:]
		if an:
			r = an.rfind('/')
			parent = an[:r] if r > 0 else ''
			pw = self.rulesets[parent]['widget']
		else:
			pw = self.host.widget

		# Save the ruleset and create widget for the anchor.
		if an not in self.rulesets:
			self.rulesets[an] = dict()
			w = fe.ltv_add_node(pw, an if an
			    else "packet filter", "rules", an)
			self.rulesets[an]['widget'] = w
		self.rulesets[an]['rules'] = o
		self.rulesets_updated += [an]

		self.pf_refcnt -= 1
		self.check_pf_refcnt_merge()

	def check_pf_refcnt_merge(self):
		fe = self.cur_frontend
		if self.pf_refcnt == 0:
			self.host.log("deleting unused anchors...")
			deleted = []
			for an in self.rulesets:
				if an in self.rulesets_updated:
					continue
				w = self.rulesets[an]['widget']
				deleted += [an]
				fe.ltv_delete_node(w)
			for an in deleted:
				del self.rulesets[an]
			self.host.log("...done!")

	def openbsd_based_on_uname_r(self, o, e=None):
		major, minor = map(int, o.split(b'.'))
		if major < 5 or (major == 5 and minor <= 4):
			self.queueing = 'altq'
		elif major == 5 and minor == 5:
			self.queueing = 'both'
		else:
			self.queueing = 'newq'
		m = "version {0}, queueing system: {1}"
		self.host.log(m.format(o.decode('ascii').strip(), self.queueing))

