
WSTALIST_PATH = "/usr/bin/wstalist"
DISCOVER_PATH = "/usr/bin/discover -j"
MCASTATUS_PATH = "/usr/bin/mca-status"
MCAFWUPDATE_PATH = "/usr/bin/mca-fwupdate"

import json

from software import QVizSoftware

class Ubiquiti(QVizSoftware):
	name = "UBNT"

	def __init__(self):
		self.stations = dict()
		self.status = dict()

	def node_types(self):
		return ['station', 'status']

	def subtrees(self):
		return [('clients', self.refresh_clients), ('status', self.async_fetch_status)]

	def refresh_clients(self, frontend, use_sudo=False):
		self.async_fetch_stations(frontend)
		self.async_fetch_discover(frontend)

	def whatis(self, path):
		for [k, h] in path:
			if k == 'station':
				t = ["station"]
				handle = self.stations[h]
			elif k == 'host':
				t = ["host"]
				handle = self.host
			else:
				self.host.log("ubnt: weird k {0} h {1}".format(k, h))
				break
		return (t, handle)

	def async_fetch_status(self, frontend, use_sudo=False):
		self.cur_frontend = frontend
		return self.host.push_task(MCASTATUS_PATH, self.finish_fetch_status)

	def finish_fetch_status(self, o, e=""):
		fe = self.cur_frontend

		if e != "":
			self.host.log(e)

		self.status = dict()
		for l in o.split(b'\r\n'):
			if l.startswith(b'deviceName='):
				first = l.split(b',')
				for x in first:
					p = x.split(b'=')
					self.status[p[0]] = p[1]
				continue
			# the rest is one thing per line...
			p = l.split(b'=')
			if len(p) == 2:
				self.status[p[0]] = p[1]


	def async_fetch_stations(self, frontend, use_sudo=False):
		self.cur_frontend = frontend
		return self.host.push_task(WSTALIST_PATH, self.finish_fetch_stations)

	def finish_fetch_stations(self, o, e=""):
		fe = self.cur_frontend

		if e != "":
			self.host.log(e)

		stations = json.loads(o.decode('ascii'))
		parent = self.host.widget
		for sta in stations:
			name = sta['mac']
			desc = sta['mac']
			if 'name' in sta:
				desc += ", " + sta['name']
			if name not in self.stations:
				self.stations[name] = sta
			else:
				for k in sta:
					self.stations[name][k] = sta[k]
			fe.ltv_add_node(parent, desc, 'station', name)

	def async_fetch_discover(self, frontend):
		self.cur_frontend = frontend
		self.host.push_task(DISCOVER_PATH, self.finish_fetch_discover)

	def finish_fetch_discover(self, o, e=""):
		fe = self.cur_frontend

		if e != "":
			self.host.log(e)

		try:
			stations = json.loads(o.decode('ascii'))['devices']
		except ValueError:
			self.host.log("invalid JSON: {0}".format(o))
			return
		parent = self.host.widget
		for sta in stations:
			name = sta['hwaddr']
			desc = sta['hwaddr']
			if 'name' in sta:
				desc += ", " + sta['hostname']
			if name not in self.stations:
				# don't save stuff from other APs
				continue
				#self.stations[name] = sta
			else:
				for k in sta:
					self.stations[name][k] = sta[k]
			fe.ltv_add_node(parent, desc, 'station', name)

	def async_upgrade_firmware(self, filename):
		from os import listdir

		allowed_fw = listdir("firmware")
		if filename not in allowed_fw:
			self.host.log("file {0} not allowed!".format(filename))
			return

		opath = "firmware/{0}".format(filename)
		npath = "/tmp/{0}".format(filename)
		self.host.scp_put(opath, npath)
		self.host.push_task("md5sum {0}".format(npath), self.finish_fw_md5_check)

	def finish_fw_md5_check(self, o, e=None):
		if e:
			msg = "md5sum cmd failed: " + e
			self.host.log(msg)
			return
		[hissum, hisname] = o.split()
		mysum = self.host.scp_md5.hexdigest()
		myname = self.host.scp_filename

		del self.host.scp_md5, self.host.scp_filename
		if myname != hisname:
			msg = "other side cocked up the names, {0} and {1}"
			self.host.log(msg.format(myname, hisname))
			return
		if mysum != hissum:
			m = "{0}: MD5 checksum not {1}, but {2}!".format(myname, mysum, hissum)
			self.host.log(m)
			self.host.push_task("rm {0}".format(myname), self.finish_rm)
			return
		self.host.log("MD5 SUM OK!")
		update_cmd = MCAFWUPDATE_PATH + " {0}".format(myname)
		self.host.push_task(update_cmd, self.finish_update_fw)

	def finish_update_fw(self, o, e=None):
		if e:
			msg = "mca-fwupdate stderr: " + e
			self.host.log(msg)
			self.host.set_status("firmware upgrade didn't go well")
		if o.startswith("Estimated upgrade time: "):
			secs = int(o.split()[3])
			self.host.log("upgrade will take {0} seconds".format(secs))
		else:
			msg = "mca-fwupdate stdout: " + o
			self.host.log(msg)
			self.host.set_status("firmware upgrade didn't go well")

	def finish_rm(self, o, e=None):
		if e:
			self.host.log(e)

